"""
Tests for plotting operations
"""

import itertools
import os
import shutil
import sys
import tempfile
import unittest
from contextlib import contextmanager
from unittest import TestCase

import numpy as np
import pandas as pd
import xarray as xr

from esa_climate_toolbox.core.op import OP_REGISTRY
from esa_climate_toolbox.core.types import ValidationError
from esa_climate_toolbox.ops.plot import plot
from esa_climate_toolbox.ops.plot import plot_line
from esa_climate_toolbox.ops.plot import plot_map
from esa_climate_toolbox.ops.plot import plot_scatter
from esa_climate_toolbox.util.misc import object_to_qualified_name

_counter = itertools.count()
ON_WIN = sys.platform == 'win32'


@contextmanager
def create_tmp_file(name, ext):
    tmp_dir = tempfile.mkdtemp()
    path = os.path.join(tmp_dir, '{}{}.{}'.format(name,
                                                  next(_counter),
                                                  ext))
    try:
        yield path
    finally:
        try:
            shutil.rmtree(tmp_dir)
        except OSError:
            if not ON_WIN:
                raise


@unittest.skipIf(condition=os.environ.get('ECT_DISABLE_PLOT_TESTS', None),
                 reason="skipped if ECT_DISABLE_PLOT_TESTS=1")
class TestPlotMap(TestCase):
    """
    Test plot_map() function
    """

    def test_plot_map(self):
        # Test the nominal functionality. This doesn't check that the plot is what
        # is expected, rather, it simply tests if it seems to have been created
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.random.rand(5, 10, 2)),
            'second': (['lat', 'lon', 'time'], np.random.rand(5, 10, 2)),
            'lat': np.linspace(-89.5, 89.5, 5),
            'lon': np.linspace(-179.5, 179.5, 10),
            'time': pd.date_range('2000-01-01', periods=2)})

        with create_tmp_file('remove_me', 'png') as tmp_file:
            plot_map(dataset, file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))

        # Test if an error is raised when an unsupported format is passed
        with create_tmp_file('remove_me', 'pgm') as tmp_file:
            with self.assertRaises(ValueError):
                plot_map(dataset, file=tmp_file)
            self.assertFalse(os.path.isfile(tmp_file))

        # Test if extents can be used
        with create_tmp_file('remove_me', 'pdf') as tmp_file:
            plot_map(dataset,
                     var='second',
                     region='-40.0, -20.0, 50.0, 60.0',
                     file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))

        # Test time slice selection
        with create_tmp_file('remove_me', 'png') as tmp_file:
            plot_map(dataset, indexers="time='2000-01-01'", file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))

    def test_plot_map_exceptions(self):
        # Test if the corner cases are detected without creating a plot for it.

        # Test value error is raised when passing an unexpected dataset type
        with create_tmp_file('remove_me', 'jpeg') as tmp_file:
            with self.assertRaises(ValueError) as cm:
                plot_map([1, 2, 4], file=tmp_file)
            self.assertEqual(str(cm.exception),
                             "Input 'ds' for operation "
                             "'esa_climate_toolbox.ops.plot.plot_map' "
                             "must be of type 'Dataset', but got type 'list'.")
            self.assertFalse(os.path.isfile(tmp_file))

        # Test the extensions bound checking
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.random.rand(2, 4, 1)),
            'lat': np.linspace(-89.5, 89.5, 2),
            'lon': np.linspace(-179.5, 179.5, 4)})

        with self.assertRaises(ValueError):
            region = '-40.0, -95.0, 50.0, 60.0',
            plot_map(dataset, region=region)

        with self.assertRaises(ValueError):
            region = '-40.0, -20.0, 50.0, 95.0',
            plot_map(dataset, region=region)

        with self.assertRaises(ValueError):
            region = '-181.0, -20.0, 50.0, 60.0',
            plot_map(dataset, region=region)

        with self.assertRaises(ValueError):
            region = '-40.0, -20.0, 181.0, 60.0',
            plot_map(dataset, region=region)

        with self.assertRaises(ValueError):
            region = '-40.0, -20.0, 50.0, -25.0',
            plot_map(dataset, region=region)

        with self.assertRaises(ValueError):
            region = '-20.0, -20.0, -25.0, 60.0',
            plot_map(dataset, region=region)

        # Test temporal slice validation
        with self.assertRaises(ValueError):
            plot_map(dataset, time=0)

    def test_registered(self):
        """
        Test nominal execution of the function as a registered operation.
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(plot_map))
        dataset = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.random.rand(5, 10, 2)),
            'second': (['lat', 'lon', 'time'], np.random.rand(5, 10, 2)),
            'lat': np.linspace(-89.5, 89.5, 5),
            'lon': np.linspace(-179.5, 179.5, 10),
            'time': pd.date_range('2000-01-01', periods=2)})

        with create_tmp_file('remove_me', 'png') as tmp_file:
            reg_op(ds=dataset, file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))


@unittest.skipIf(condition=os.environ.get('ECT_DISABLE_PLOT_TESTS', None),
                 reason="skipped if ECT_DISABLE_PLOT_TESTS=1")
class TestPlot(TestCase):
    """
    Test plot() function
    """

    def test_plot(self):
        # Test plot
        dataset = xr.Dataset({
            'first': np.random.rand(10)})

        with create_tmp_file('remove_me', 'jpg') as tmp_file:
            plot(dataset, 'first', file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))

    def test_registered(self):
        """
        Test nominal execution of the function as a registered operation.
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(plot))
        # Test plot
        dataset = xr.Dataset({
            'first': np.random.rand(10)})

        with create_tmp_file('remove_me', 'jpg') as tmp_file:
            reg_op(ds=dataset, var='first', file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))


@unittest.skipIf(condition=os.environ.get('ECT_DISABLE_PLOT_TESTS', None),
                 reason="skipped if ECT_DISABLE_PLOT_TESTS=1")
class TestPlotLine(TestCase):
    """
    Test plot_line() function
    """

    def test_plot(self):
        # Test plot_line
        single_dim_ds = xr.Dataset({
            'first': (['time'], np.random.rand(10)),
            'second': (['time'], np.random.rand(10)),
            'time': pd.date_range('2000-01-01', periods=10)})

        # Test with only 1 variable selected
        with create_tmp_file('remove_me', 'jpg') as tmp_file:
            plot_line(single_dim_ds, ['first'], file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))

        # Test with 2 variables selected
        with create_tmp_file('remove_me', 'jpg') as tmp_file:
            plot_line(single_dim_ds, ['first', 'second'], file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))

        # Test with specified formats for each variable
        with create_tmp_file('remove_me', 'jpg') as tmp_file:
            plot_line(single_dim_ds, ['first', 'second'], fmt='r^--;b-', file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))

        # Test with partially specified formats
        with create_tmp_file('remove_me', 'jpg') as tmp_file:
            plot_line(single_dim_ds, ['first', 'second'], fmt='r^-', file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))

        multi_dim_ds = xr.Dataset({
            'first': (['lat', 'lon', 'layers', 'time'], np.random.rand(5, 10, 12, 2)),
            'second': (['lat', 'lon', 'layers', 'time'], np.random.rand(5, 10, 12, 2)),
            'lat': np.linspace(-89.5, 89.5, 5),
            'lon': np.linspace(-179.5, 179.5, 10),
            'layers': np.linspace(0, 12, 12),
            'time': pd.date_range('2000-01-01', periods=2)})

        # Test value error is raised when there are too many dimensions to plot
        with create_tmp_file('remove_me', 'jpeg') as tmp_file:
            with self.assertRaises(ValueError) as cm:
                plot_line(multi_dim_ds, ['first', 'second'], file=tmp_file)
            self.assertEqual(
                "Unable to plot because variable first has more than one dimension: "
                "('lat', 'lon', 'layers', 'time'). "
                "To specify value(s) of these dimension(s), please use the indexers.",
                str(cm.exception)
            )
            self.assertFalse(os.path.isfile(tmp_file))

        # Now with indexers but there are still some unspecified dimension
        with create_tmp_file('remove_me', 'jpeg') as tmp_file:
            with self.assertRaises(ValueError) as cm:
                plot_line(multi_dim_ds, ['first', 'second'],
                          indexers="lat=89.5, lon=-179.5",
                          file=tmp_file
                          )
            self.assertEqual(
                "Unable to plot because variable first has more than one dimension: "
                "('layers', 'time'). "
                "To specify value(s) of these dimension(s), please use the indexers.",
                str(cm.exception)
            )
            self.assertFalse(os.path.isfile(tmp_file))

        # Now with indexers with all dimensions (except 1) specified
        with create_tmp_file('remove_me', 'jpeg') as tmp_file:
            plot_line(multi_dim_ds, ['first', 'second'],
                      indexers="lat=89.5, lon=-179.5, layers=0",
                      file=tmp_file
                      )
            self.assertTrue(os.path.isfile(tmp_file))

        # Now with indexers with all dimensions (except 1) and label specified
        with create_tmp_file('remove_me', 'jpeg') as tmp_file:
            plot_line(multi_dim_ds,
                      ['first', 'second'],
                      label='time',
                      indexers="lat=89.5, lon=-179.5, layers=0",
                      file=tmp_file
                      )
            self.assertTrue(os.path.isfile(tmp_file))

        # Now with indexers and label specified with some overlapping
        with create_tmp_file('remove_me', 'jpeg') as tmp_file:
            with self.assertRaises(ValueError) as cm:
                plot_line(multi_dim_ds,
                          ['first', 'second'],
                          label='layers',
                          indexers="lat=89.5, lon=-179.5, layers=0",
                          file=tmp_file
                          )
            self.assertEqual(
                "Dimension 'layers' is also specified as indexers. Please ensure "
                "that a dimension is used exclusively either as indexers "
                "or as the selected dimension.",
                str(cm.exception)
            )
            self.assertFalse(os.path.isfile(tmp_file))

    def test_registered(self):
        """
        Test nominal execution of the function as a registered operation.
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(plot_line))
        # Test plot
        dataset = xr.Dataset({
            'first': (['time'], np.random.rand(10)),
            'second': (['time'], np.random.rand(10)),
            'time': pd.date_range('2000-01-01', periods=10)})

        with create_tmp_file('remove_me', 'jpg') as tmp_file:
            reg_op(ds=dataset, var_names=['first', 'second'], file=tmp_file)
            self.assertTrue(os.path.isfile(tmp_file))


@unittest.skipIf(condition=os.environ.get('ECT_DISABLE_PLOT_TESTS', None),
                 reason="skipped if ECT_DISABLE_PLOT_TESTS=1")
class TestPlotScatter(TestCase):
    """
    Test plot_scatter() function
    """

    width = 360 - 1
    height = 180 - 1
    num_time_steps = 1
    # Sea Level MSL Dummy
    ds1 = xr.Dataset({
        'local_msl_trend': (
            ['time', 'lat', 'lon'], np.random.rand(num_time_steps, height, width)
        ),
        'local_msl_trend_error': (
            ['time', 'lat', 'lon'], np.random.rand(num_time_steps, height, width)
        ),
        'lon': np.linspace(-179.5, 179.5, width),
        'lat': np.linspace(-89.5, 89.5, height),
        'time': pd.date_range('2000-01-01', periods=num_time_steps, freq='D')
    })

    periods = [0.5, 10.0]
    num_periods = len(periods)
    # Sea Level MSLAMPH Dummy
    ds2 = xr.Dataset({
        'ampl': (
            ['time', 'period', 'lat', 'lon'],
            10 * np.random.rand(num_time_steps, num_periods, height, width)
        ),
        'phase': (
            ['time', 'period', 'lat', 'lon'],
            360 * np.random.rand(num_time_steps, num_periods, height, width)
        ),
        'lon': np.linspace(-179.5, 179.5, width),
        'lat': np.linspace(-89.5, 89.5, height),
        'period': np.array(periods),
        'time': pd.date_range('2000-01-01', periods=num_time_steps, freq='D')
    })

    def test_all_types_work_fine(self):
        for type in ('Point', 'Hexbin', '2D Histogram'):
            with create_tmp_file('remove_me', 'png') as tmp_file:
                plot_scatter(ds1=self.ds1,
                             ds2=self.ds2,
                             var1='local_msl_trend',
                             var2='ampl',
                             indexers1=None,
                             indexers2=dict(period=0.5),
                             type=type,
                             title='Local MSL vs Amplitude',
                             properties=None,
                             file=tmp_file)
                self.assertTrue(os.path.isfile(tmp_file))

    def test_illegal_type(self):
        with self.assertRaises(ValidationError) as cm:
            plot_scatter(ds1=self.ds1,
                         ds2=self.ds2,
                         var1='local_msl_trend',
                         var2='ampl',
                         indexers1=None,
                         indexers2=dict(period=0.5),
                         type='2D Density')
        self.assertEqual("Input 'type' for operation "
                         "'esa_climate_toolbox.ops.plot.plot_scatter'"
                         " must be one of ['Point', 'Hexbin', '2D Histogram'].",
                         f'{cm.exception}')

    def test_illegal_var(self):
        with self.assertRaises(ValidationError) as cm:
            plot_scatter(ds1=self.ds1,
                         ds2=self.ds2,
                         var1='global_msl_trend',
                         var2='ampl',
                         indexers1=None,
                         indexers2=dict(period=0.5),
                         type='Point')
        self.assertEqual(
            '"global_msl_trend" is not a variable in dataset given by "ds1"',
            f'{cm.exception}'
        )

    def test_illegal_no_common_dim(self):
        with self.assertRaises(ValidationError) as cm:
            plot_scatter(ds1=self.ds1,
                         ds2=self.ds2,
                         var1='local_msl_trend',
                         var2='period',
                         indexers1=None,
                         indexers2=dict(period=0.5),
                         type='Point')
        self.assertTrue(f'{cm.exception}'.startswith(
            '"var1" and "var2" have no dimensions in common: ')
        )

    def test_illegal_no_common_dims(self):
        with self.assertRaises(ValidationError) as cm:
            plot_scatter(ds1=self.ds1,
                         ds2=self.ds2,
                         var1='local_msl_trend',
                         var2='phase',
                         indexers1=None,
                         indexers2=None,
                         type='Point')
        self.assertTrue(f'{cm.exception}'.startswith(
            'Remaining dimensions of data from "var1" must be equal to '
            'remaining dimensions of data from "var2", but ')
        )



from collections import namedtuple
from datetime import datetime, date
from io import StringIO
from typing import Union, Tuple
from unittest import TestCase

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely.wkt
import xarray as xr
from shapely.geometry import Point, Polygon

from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.op import OpRegistry
from esa_climate_toolbox.core.types import Arbitrary
from esa_climate_toolbox.core.types import DatasetLike
from esa_climate_toolbox.core.types import DataFrameLike
from esa_climate_toolbox.core.types import DictLike
from esa_climate_toolbox.core.types import DimName
from esa_climate_toolbox.core.types import DimNamesLike
from esa_climate_toolbox.core.types import FileLike
from esa_climate_toolbox.core.types import GeoDataFrame
from esa_climate_toolbox.core.types import GeometryLike
from esa_climate_toolbox.core.types import HTML
from esa_climate_toolbox.core.types import HTMLLike
from esa_climate_toolbox.core.types import Like
from esa_climate_toolbox.core.types import Literal
from esa_climate_toolbox.core.types import PointLike
from esa_climate_toolbox.core.types import PolygonLike
from esa_climate_toolbox.core.types import TimeLike
from esa_climate_toolbox.core.types import TimeRangeLike
from esa_climate_toolbox.core.types import ValidationError
from esa_climate_toolbox.core.types import VarName
from esa_climate_toolbox.core.types import VarNamesLike
from esa_climate_toolbox.util.misc import object_to_qualified_name
from esa_climate_toolbox.util.misc import OrderedDict

# 'ExamplePoint' is an example type which may come from
# the ESA Climate Toolbox API or other required API.
ExamplePoint = namedtuple('ExamplePoint', ['x', 'y'])


# 'ExampleType' represents a type to be used for values that should actually
# be a 'ExamplePoint' but also have a representation as a `str`
# such as "2.1,4.3", ar a tuple of two floats (2.1,4.3).
# The "typing type" for this is given by ExampleType.TYPE.

class ExampleType(Like[ExamplePoint]):

    TYPE = Union[ExamplePoint, Tuple[float, float], str]

    @classmethod
    def convert(cls, value, default=None) -> ExamplePoint:
        try:
            if isinstance(value, ExamplePoint):
                return value
            if isinstance(value, str):
                pair = value.split(',')
                return ExamplePoint(float(pair[0]), float(pair[1]))
            return ExamplePoint(value[0], value[1])
        except Exception:
            raise ValidationError('Cannot convert value <%s> to %s.' %
                                  (repr(value), cls.name()))

    @classmethod
    def format(cls, value: ExamplePoint) -> str:
        return "%s, %s" % (value.x, value.y)


# TestType = NewType('TestType', _TestType)


# 'scale_point' is an example operation that makes use of the TestType type
# for argument point_like

_OP_REGISTRY = OpRegistry()


@op_input("point_like", data_type=ExampleType, registry=_OP_REGISTRY)
def scale_point(point_like: ExampleType.TYPE, factor: float) -> ExamplePoint:
    point = ExampleType.convert(point_like)
    return ExamplePoint(factor * point.x, factor * point.y)


class ExampleTypeTest(TestCase):

    def test_use(self):
        self.assertEqual(scale_point("2.4, 4.8", 0.5), ExamplePoint(1.2, 2.4))
        self.assertEqual(scale_point((2.4, 4.8), 0.5), ExamplePoint(1.2, 2.4))
        self.assertEqual(scale_point(ExamplePoint(2.4, 4.8), 0.5), ExamplePoint(1.2, 2.4))

    def test_abuse(self):
        with self.assertRaises(ValidationError) as cm:
            scale_point("A, 4.8", 0.5)
        self.assertEqual(str(cm.exception),
                         "Input 'point_like' for operation "
                         "'test.core.test_types.scale_point': "
                         "Cannot convert value <'A, 4.8'> to ExampleType.")

        with self.assertRaises(ValidationError) as cm:
            scale_point(25.1, 0.5)
        self.assertEqual(str(cm.exception),
                         "Input 'point_like' for operation "
                         "'test.core.test_types.scale_point': "
                         "Cannot convert value <25.1> to ExampleType.")

    def test_registered_op(self):
        registered_op = _OP_REGISTRY.get_op(
            object_to_qualified_name(scale_point)
        )
        point = registered_op(point_like="2.4, 4.8", factor=0.5)
        self.assertEqual(point, ExamplePoint(1.2, 2.4))

    def test_name(self):
        self.assertEqual(ExampleType.name(), "ExampleType")

    def test_accepts(self):
        self.assertTrue(ExampleType.accepts("2.4, 4.8"))
        self.assertTrue(ExampleType.accepts((2.4, 4.8)))
        self.assertTrue(ExampleType.accepts([2.4, 4.8]))
        self.assertTrue(ExampleType.accepts(ExamplePoint(2.4, 4.8)))
        self.assertFalse(ExampleType.accepts("A, 4.8"))
        self.assertFalse(ExampleType.accepts(25.1))

    def test_format(self):
        self.assertEqual(ExampleType.format(ExamplePoint(2.4, 4.8)), "2.4, 4.8")


class HTMLLikeTest(TestCase):
    """
    Test the HTMLLike type
    """

    def test_accepts(self):
        self.assertTrue(HTMLLike.accepts('abc'))
        self.assertTrue(HTMLLike.accepts(HTML('abc')))
        self.assertTrue(HTMLLike.accepts(42))

    def test_convert(self):
        actual = HTMLLike.convert('abc')
        self.assertIsInstance(actual, HTML)
        self.assertEqual(actual, 'abc')

    def test_format(self):
        actual = VarNamesLike.format(HTML('abc'))
        self.assertIsInstance(actual, str)
        self.assertEqual(actual, 'abc')


class VarNamesLikeTest(TestCase):
    """
    Test the VarNamesLike type
    """

    def test_accepts(self):
        self.assertTrue(VarNamesLike.accepts('aa'))
        self.assertTrue(VarNamesLike.accepts('aa,bb,cc'))
        self.assertTrue(VarNamesLike.accepts(['aa', 'bb', 'cc']))
        self.assertFalse(VarNamesLike.accepts(1.0))
        self.assertFalse(VarNamesLike.accepts([1, 2, 4]))
        self.assertFalse(VarNamesLike.accepts(['aa', 2, 'bb']))

    def test_convert(self):
        expected = ['aa', 'b*', 'cc']
        actual = VarNamesLike.convert('aa,b*,cc')
        self.assertEqual(actual, expected)

        with self.assertRaises(ValidationError) as err:
            VarNamesLike.convert(['aa', 1, 'bb'])
        self.assertEqual(
            str(err.exception), 'List of variables names expected.'
        )
        self.assertEqual(None, VarNamesLike.convert(None))

    def test_format(self):
        self.assertEqual(VarNamesLike.format(['aa', 'bb', 'cc']), "aa, bb, cc")
        self.assertEqual(VarNamesLike.format(['aa']), "aa")
        self.assertEqual(VarNamesLike.format([]), "")
        self.assertEqual(VarNamesLike.format(None), "")


class VarNameTest(TestCase):
    """
    Test the VarName type
    """

    def test_accepts(self):
        self.assertTrue(VarName.accepts('aa'))
        self.assertFalse(VarName.accepts(['aa', 'bb', 'cc']))
        self.assertFalse(VarName.accepts(1.0))

    def test_convert(self):
        expected = 'aa'
        actual = VarName.convert('aa')
        self.assertEqual(actual, expected)

        with self.assertRaises(ValidationError) as err:
            VarName.convert(['aa', 'bb', 'cc'])
        self.assertEqual(str(err.exception), 'Variable name expected.')
        self.assertEqual(None, VarName.convert(None))

    def test_format(self):
        self.assertEqual('aa', VarName.format('aa'))


class DimNamesLikeTest(TestCase):
    """
    Test the DimNamesLike type
    """

    def test_accepts(self):
        self.assertTrue(DimNamesLike.accepts('aa'))
        self.assertTrue(DimNamesLike.accepts('aa,bb,cc'))
        self.assertTrue(DimNamesLike.accepts(['aa', 'bb', 'cc']))
        self.assertFalse(DimNamesLike.accepts(1.0))
        self.assertFalse(DimNamesLike.accepts([1, 2, 4]))
        self.assertFalse(DimNamesLike.accepts(['aa', 2, 'bb']))

    def test_convert(self):
        expected = ['aa', 'b*', 'cc']
        actual = DimNamesLike.convert('aa,b*,cc')
        self.assertEqual(actual, expected)

        with self.assertRaises(ValidationError) as err:
            DimNamesLike.convert(['aa', 1, 'bb'])
        self.assertEqual(
            str(err.exception), 'List of dimension names expected.'
        )
        self.assertEqual(None, DimNamesLike.convert(None))

    def test_format(self):
        self.assertEqual(DimNamesLike.format(['aa', 'bb', 'cc']), "aa, bb, cc")
        self.assertEqual(DimNamesLike.format(['aa']), "aa")
        self.assertEqual(DimNamesLike.format([]), "")
        self.assertEqual(DimNamesLike.format(None), "")


class DimNameTest(TestCase):
    """
    Test the DimName type
    """

    def test_accepts(self):
        self.assertTrue(DimName.accepts('aa'))
        self.assertFalse(DimName.accepts(['aa', 'bb', 'cc']))
        self.assertFalse(DimName.accepts(1.0))

    def test_convert(self):
        expected = 'aa'
        actual = DimName.convert('aa')
        self.assertEqual(actual, expected)

        with self.assertRaises(ValidationError) as err:
            DimName.convert(['aa', 'bb', 'cc'])
        self.assertEqual(str(err.exception), 'Dimension name expected.')
        self.assertEqual(None, DimName.convert(None))

    def test_format(self):
        self.assertEqual('aa', DimName.format('aa'))


class FileLikeTest(TestCase):
    """
    Test the FileLike type
    """

    def test_accepts(self):
        self.assertTrue(FileLike.accepts(None))
        self.assertTrue(FileLike.accepts(''))
        self.assertTrue(FileLike.accepts('a/b/c'))
        self.assertTrue(FileLike.accepts(StringIO()))
        self.assertFalse(FileLike.accepts(2))
        self.assertFalse(FileLike.accepts(True))

    def test_convert(self):
        self.assertEqual(FileLike.convert(None), None)
        self.assertEqual(FileLike.convert(''), None)
        self.assertEqual(FileLike.convert('a/b/c'), 'a/b/c')
        io = StringIO()
        self.assertEqual(FileLike.convert(io), io)

    def test_format(self):
        self.assertEqual(FileLike.format(None), '')
        self.assertEqual(FileLike.format('a/b/c'), 'a/b/c')
        io = StringIO()
        self.assertEqual(FileLike.format(io), '')


class DictLikeTest(TestCase):
    """
    Test the DictLike type
    """

    def test_accepts(self):
        self.assertTrue(DictLike.accepts(None))
        self.assertTrue(DictLike.accepts(''))
        self.assertTrue(DictLike.accepts('  '))
        self.assertTrue(DictLike.accepts('a=6, b=5.3, c=True, d="Hello"'))

        self.assertFalse(DictLike.accepts('{a=True}'))
        self.assertFalse(DictLike.accepts('a=true'))
        self.assertFalse(DictLike.accepts('{a, b}'))
        self.assertFalse(DictLike.accepts(['aa', 'bb', 'cc']))
        self.assertFalse(DictLike.accepts(1.0))

    def test_convert(self):
        self.assertEqual(DictLike.convert(None), None)
        self.assertEqual(DictLike.convert(''), None)
        self.assertEqual(DictLike.convert('  '), None)
        self.assertEqual(
            DictLike.convert('name="bibo", thres=0.5, drop=False'),
            dict(name="bibo", thres=0.5, drop=False)
        )

        with self.assertRaises(ValidationError) as err:
            DictLike.convert('{a=8, b}')
        self.assertEqual(
            str(err.exception),
            "Value '{a=8, b}' cannot be converted into a 'DictLike'."
        )

    def test_format(self):
        self.assertEqual(
            DictLike.format(OrderedDict(
                [('name', 'bibo'), ('thres', 0.5), ('drop', True)])
            ),
            "name='bibo', thres=0.5, drop=True"
        )

    def test_to_json(self):
        self.assertEqual(
            DictLike.to_json(OrderedDict(
                [('name', 'bibo'), ('thres', 0.5), ('drop', True)]
            )),
            "name='bibo', thres=0.5, drop=True"
        )

    def test_from_json(self):
        self.assertEqual(
            DictLike.from_json("name='bibo', thres=0.5, drop=True"),
            dict(name='bibo', thres=0.5, drop=True)
        )


class PointLikeTest(TestCase):
    """
    Test the PointLike type
    """

    def test_accepts(self):
        self.assertTrue(PointLike.accepts(""))
        self.assertTrue(PointLike.accepts("\t\n "))
        self.assertTrue(PointLike.accepts("2.4, 4.8\n"))
        self.assertTrue(PointLike.accepts((2.4, 4.8)))
        self.assertTrue(PointLike.accepts([2.4, 4.8]))
        self.assertTrue(PointLike.accepts(Point(2.4, 4.8)))
        self.assertTrue(PointLike.accepts(Point(2.4, 4.8).wkt))
        self.assertFalse(PointLike.accepts("A, 4.8"))
        self.assertFalse(PointLike.accepts(25.1))

    def test_convert(self):
        self.assertEqual(PointLike.convert(None), None)
        self.assertEqual(PointLike.convert(''), None)
        self.assertEqual(PointLike.convert('0.0,1.0'), Point(0.0, 1.0))
        with self.assertRaises(ValidationError) as err:
            PointLike.convert('0.0,abc')
        self.assertEqual(str(err.exception),
                         "Value cannot be converted into a 'PointLike': "
                         "Invalid geometry WKT format."
                         )
        self.assertEqual(PointLike.convert('POINT(0.0 1.0)'), Point(0.0, 1.0))

    def test_format(self):
        self.assertEqual(PointLike.format(Point(2.4, 4.8)), "2.4, 4.8")


class PolygonLikeTest(TestCase):
    """
    Test the PolygonLike type
    """

    def test_accepts(self):
        self.assertTrue(PolygonLike.accepts(""))
        self.assertTrue(PolygonLike.accepts(" \t"))
        self.assertTrue(PolygonLike.accepts("0.0,0.0,1.1,1.1"))
        self.assertTrue(PolygonLike.accepts("0.0, 0.0, 1.1, 1.1"))

        coords = [(10.4, 20.2), (30.8, 20.2), (30.8, 40.8), (10.4, 40.8)]
        pol = Polygon(coords)
        self.assertTrue(PolygonLike.accepts(coords))
        self.assertTrue(PolygonLike.accepts(pol))
        self.assertTrue(PolygonLike.accepts(pol.wkt))
        self.assertTrue(PolygonLike.accepts(pol.bounds))

        self.assertFalse(PolygonLike.accepts("0.0,aaa,1.1,1.1"))
        self.assertFalse(PolygonLike.accepts("0.0, aaa, 1.1, 1.1"))
        self.assertFalse(PolygonLike.accepts(
            [(0.0, 0.0), (0.0, 1.0), (1.0, 'aaa'), (1.0, 0.0)]
        ))
        self.assertFalse(PolygonLike.accepts(
            [(0.0, 0.0), (0.0, 1.0), 'Guten Morgen, Berlin!', (1.0, 0.0)])
        )
        self.assertFalse(PolygonLike.accepts(
            Polygon([(0.0, 0.0), (0.0, 0.0), (0.0, 0.0), (0.0, 0.0)]))
        )
        self.assertFalse(PolygonLike.accepts('MULTIPOLYGON()'))
        self.assertFalse(PolygonLike.accepts('Something_in_the_month_of_May'))
        self.assertFalse(PolygonLike.accepts(1.0))

    def test_convert(self):
        self.assertEqual(PolygonLike.convert(None), None)
        self.assertEqual(PolygonLike.convert(''), None)
        coords = [(10.4, 20.2), (30.8, 20.2), (30.8, 40.8), (10.4, 40.8)]
        self.assertTrue(PolygonLike.convert(coords), Polygon(coords))
        self.assertTrue(
            PolygonLike.convert([10.4, 20.2, 30.8, 40.8]), Polygon(coords)
        )

        with self.assertRaises(ValidationError) as err:
            PolygonLike.convert('aaa')
        self.assertEqual(str(err.exception),
                         "Value cannot be converted into a 'PolygonLike': "
                         "Invalid geometry WKT format.")

    def test_format(self):
        self.assertEqual(PolygonLike.format(None), '')
        coords = [(10.4, 20.2), (30.8, 20.2), (30.8, 40.8), (10.4, 40.8)]
        pol = PolygonLike.convert(coords)
        self.assertEqual(
            PolygonLike.format(pol),
            'POLYGON ((10.4 20.2, 30.8 20.2, 30.8 40.8, 10.4 40.8, 10.4 20.2))'
        )

    def test_json(self):
        self.assertEqual(
            PolygonLike.from_json("-10, -10, 10, 10"), "-10, -10, 10, 10"
        )


class GeometryLikeTest(TestCase):

    def test_accepts(self):
        self.assertTrue(GeometryLike.accepts("10, 10"))
        self.assertTrue(GeometryLike.accepts("10, 10, 20, 20"))
        self.assertTrue(GeometryLike.accepts(
            [(10, 10), (20, 10), (20, 20), (10, 20), (10, 10)]
        ))
        self.assertTrue(
            GeometryLike.accepts(shapely.wkt.loads("POINT (10 10)"))
        )
        self.assertTrue(
            GeometryLike.accepts(shapely.wkt.loads("POINT (10 10)"))
        )
        self.assertTrue(
            GeometryLike.accepts(shapely.wkt.loads(
                "MULTIPOINT (10 10, 20 10, 20 20, 10 20, 10 10)"))
        )
        self.assertTrue(GeometryLike.accepts(shapely.wkt.loads(
            "LINESTRING (10 10, 20 10, 20 20, 10 20, 10 10)"))
        )
        self.assertTrue(GeometryLike.accepts(shapely.wkt.loads(
            "MULTILINESTRING ((10 10, 20 10, 20 20, 10 20, 10 10))"))
        )
        self.assertTrue(GeometryLike.accepts(shapely.wkt.loads(
            "POLYGON ((10 10, 20 10, 20 20, 10 20, 10 10))"))
        )
        self.assertTrue(GeometryLike.accepts(shapely.wkt.loads(
            "MULTIPOLYGON (((10 10, 20 10, 20 20, 10 20, 10 10)))"))
        )
        self.assertTrue(GeometryLike.accepts(shapely.wkt.loads(
            "GEOMETRYCOLLECTION (POINT (10 10), "
            "POLYGON ((10 10, 20 10, 20 20, 10 20, 10 10)))"))
        )
        self.assertTrue(GeometryLike.accepts(
            [(0.0, 0.0), (0.0, 0.0), (0.0, 0.0), (0.0, 0.0)])
        )

        self.assertFalse(GeometryLike.accepts("0.0,aaa,1.1,1.1"))
        self.assertFalse(GeometryLike.accepts("0.0, aaa, 1.1, 1.1"))

        self.assertFalse(GeometryLike.accepts('MULTIPOLYGON()'))
        self.assertFalse(GeometryLike.accepts('Something_in_the_month_of_May'))
        self.assertFalse(GeometryLike.accepts(1.0))

    def test_convert(self):
        self.assertEqual(GeometryLike.convert(None), None)
        self.assertEqual(GeometryLike.convert(""), None)
        coords = [(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)]
        self.assertTrue(GeometryLike.convert(coords), Polygon(coords))

        with self.assertRaises(ValidationError) as err:
            GeometryLike.convert('aaa')
        self.assertEqual(str(err.exception),
                         "Value cannot be converted into a 'GeometryLike': "
                         "Invalid geometry WKT format.")

    def test_format(self):
        pol = GeometryLike.convert(
            [(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)]
        )
        self.assertTrue(
            GeometryLike.format(pol), 'POLYGON ((0 0, 0 1, 1 1, 1 0, 0 0))'
        )


class TimeLikeTest(TestCase):
    """
    Test the TimeLike type
    """

    def test_accepts(self):
        self.assertTrue(TimeLike.accepts(None))
        self.assertTrue(TimeLike.accepts(''))
        self.assertTrue(TimeLike.accepts(' '))
        self.assertTrue(TimeLike.accepts('2001-01-01'))
        self.assertTrue(TimeLike.accepts(datetime(2001, 1, 1)))
        self.assertTrue(TimeLike.accepts(date(2001, 1, 1)))

        self.assertFalse(TimeLike.accepts('4.3'))
        self.assertFalse(TimeLike.accepts('2001-01-01,2001-02-01,'))
        self.assertFalse(TimeLike.accepts([datetime(2001, 1, 1)]))

    def test_convert(self):
        self.assertEqual(
            TimeLike.convert('2017-04-19'), datetime(2017, 4, 19)
        )
        self.assertEqual(
            TimeLike.convert(datetime(2017, 4, 19)), datetime(2017, 4, 19)
        )
        self.assertEqual(
            TimeLike.convert(date(2017, 4, 19)), datetime(2017, 4, 19, 12)
        )
        self.assertEqual(TimeLike.convert('  '), None)
        self.assertEqual(TimeLike.convert(None), None)

    def test_format(self):
        self.assertEqual(TimeLike.format(None), '')
        self.assertEqual(TimeLike.format(datetime(2017, 4, 19)), '2017-04-19')

    def test_json(self):
        self.assertEqual(TimeLike.to_json(datetime(2017, 4, 19)), '2017-04-19')
        self.assertEqual(
            TimeLike.from_json('2017-04-19'), datetime(2017, 4, 19)
        )


class TimeRangeLikeTest(TestCase):
    """
    Test the TimeRangeLike type
    """

    def test_accepts(self):
        self.assertTrue(TimeRangeLike.accepts(('2001-01-01', '2002-02-01')))
        self.assertTrue(
            TimeRangeLike.accepts((datetime(2001, 1, 1), datetime(2002, 2, 1)))
        )
        self.assertTrue(
            TimeRangeLike.accepts((date(2001, 1, 1), date(2002, 1, 1)))
        )
        self.assertTrue(TimeRangeLike.accepts('2001-01-01,2002-02-01'))
        self.assertTrue(TimeRangeLike.accepts('2001-01-01, 2002-02-01'))

        self.assertFalse(TimeRangeLike.accepts('2001-01-01'))
        self.assertFalse(TimeRangeLike.accepts([datetime(2001, 1, 1)]))
        self.assertFalse(TimeRangeLike.accepts('2002-01-01, 2001-01-01'))

    def test_convert(self):
        self.assertEqual(TimeRangeLike.convert(None), None)
        self.assertEqual(TimeRangeLike.convert((None, None)), None)
        self.assertEqual(TimeRangeLike.convert([None, None]), None)
        self.assertEqual(TimeRangeLike.convert(''), None)
        self.assertEqual(
            TimeRangeLike.convert((datetime(2001, 1, 1), datetime(2002, 2, 1))),
            (datetime(2001, 1, 1), datetime(2002, 2, 1))
        )
        self.assertEqual(
            TimeRangeLike.convert([datetime(2001, 1, 1), datetime(2002, 2, 1)]),
            (datetime(2001, 1, 1), datetime(2002, 2, 1))
        )
        self.assertEqual(
            TimeRangeLike.convert('2001-01-01, 2002-01-01'),
            (datetime(2001, 1, 1), datetime(2002, 1, 1, 23, 59, 59))
        )
        self.assertEqual(
            TimeRangeLike.convert('2001-01-01, 2002-01-01'),
            (datetime(2001, 1, 1), datetime(2002, 1, 1, 23, 59, 59))
        )

        with self.assertRaises(ValidationError) as err:
            TimeRangeLike.convert('2002-01-01, 2001-01-01')
        self.assertTrue('cannot be converted into a' in str(err.exception))

    def test_format(self):
        self.assertEqual(TimeRangeLike.format(None), '')
        self.assertEqual(
            TimeRangeLike.format((datetime(2001, 1, 1), datetime(2002, 1, 1))),
            '2001-01-01, 2002-01-01'
        )
        self.assertEqual(
            TimeRangeLike.format((datetime(2001, 1, 1, 12),
                                  datetime(2002, 1, 1, 9, 30, 2))),
            '2001-01-01T12:00:00, 2002-01-01T09:30:02'
        )


class TypeNamesTest(TestCase):
    """
    This test fails, if any of the expected type names change.
    """

    def test_python_primitive_type_names(self):
        """
        Python primitive types
        """
        self.assertEqual(object_to_qualified_name(bool), 'bool')
        self.assertEqual(object_to_qualified_name(int), 'int')
        self.assertEqual(object_to_qualified_name(float), 'float')
        self.assertEqual(object_to_qualified_name(str), 'str')

    def test_cdm_type_names(self):
        """
        Common Data Model (CDM) types
        """
        self.assertEqual(object_to_qualified_name(np.ndarray), 'numpy.ndarray')
        self.assertEqual(object_to_qualified_name(xr.Dataset),
                         'xarray.core.dataset.Dataset')
        self.assertEqual(object_to_qualified_name(xr.DataArray),
                         'xarray.core.dataarray.DataArray')
        self.assertEqual(object_to_qualified_name(gpd.GeoDataFrame),
                         'geopandas.geodataframe.GeoDataFrame')
        self.assertEqual(object_to_qualified_name(gpd.GeoSeries),
                         'geopandas.geoseries.GeoSeries')
        self.assertEqual(object_to_qualified_name(pd.DataFrame),
                         'pandas.core.frame.DataFrame')
        self.assertEqual(object_to_qualified_name(pd.Series),
                         'pandas.core.series.Series')

    def test_op_api_type_names(self):
        """
        Additional toolbox types used by operations API.
        """
        self.assertEqual(object_to_qualified_name(VarName),
                         'esa_climate_toolbox.core.types.VarName')
        self.assertEqual(object_to_qualified_name(VarNamesLike),
                         'esa_climate_toolbox.core.types.VarNamesLike')
        self.assertEqual(object_to_qualified_name(PointLike),
                         'esa_climate_toolbox.core.types.PointLike')
        self.assertEqual(object_to_qualified_name(PolygonLike),
                         'esa_climate_toolbox.core.types.PolygonLike')
        self.assertEqual(object_to_qualified_name(GeometryLike),
                         'esa_climate_toolbox.core.types.GeometryLike')
        self.assertEqual(object_to_qualified_name(TimeRangeLike),
                         'esa_climate_toolbox.core.types.TimeRangeLike')


class ArbitraryTest(TestCase):
    def test_convert(self):
        self.assertEqual(Arbitrary.convert(None), None)
        self.assertEqual(Arbitrary.convert(434), 434)
        self.assertEqual(Arbitrary.convert(3.4), 3.4)
        self.assertEqual(Arbitrary.convert(True), True)
        self.assertEqual(Arbitrary.convert((3, 5, 7)), (3, 5, 7))
        self.assertEqual(Arbitrary.convert('abc'), 'abc')

    def test_format(self):
        self.assertEqual(Arbitrary.format(None), '')
        self.assertEqual(Arbitrary.format(434), '434')
        self.assertEqual(Arbitrary.format(3.4), '3.4')
        self.assertEqual(Arbitrary.format("abc"), "abc")
        self.assertEqual(Arbitrary.format(True), 'True')


class LiteralTest(TestCase):
    def test_convert(self):
        self.assertEqual(Literal.convert(''), None)
        self.assertEqual(Literal.convert('None'), None)
        self.assertEqual(Literal.convert('434'), 434)
        self.assertEqual(Literal.convert('3.4'), 3.4)
        self.assertEqual(Literal.convert('True'), True)
        self.assertEqual(Literal.convert('"abc"'), 'abc')
        self.assertEqual(Literal.convert('[3, 5, 7]'), [3, 5, 7])
        self.assertEqual(Literal.convert('(3, 5, 7)'), (3, 5, 7))

        with self.assertRaises(ValidationError):
            Literal.convert('[1,2')
        with self.assertRaises(ValidationError):
            Literal.convert('abc')

    def test_format(self):
        self.assertEqual(Literal.format(None), '')
        self.assertEqual(Literal.format(434), '434')
        self.assertEqual(Literal.format(3.4), '3.4')
        self.assertEqual(Literal.format("abc"), "'abc'")
        self.assertEqual(Literal.format(True), 'True')
        self.assertEqual(Literal.format([1, 2, 3]), '[1, 2, 3]')


class DatasetLikeTest(TestCase):
    def test_convert(self):
        self.assertEqual(DatasetLike.convert(None), None)

        data = {'time': ['2000-01-01', '2000-01-02', '2000-01-03'],
                'c1': [4, 5, 6],
                'c2': [6, 7, 8]}
        pd_ds = pd.DataFrame(data=data)
        pd_ds = pd_ds.set_index('time')
        pd_ds.index = pd.to_datetime(pd_ds.index)
        xr_ds = xr.Dataset(data_vars=data)
        self.assertIsInstance(DatasetLike.convert(xr_ds), xr.Dataset)
        self.assertIsInstance(DatasetLike.convert(pd_ds), xr.Dataset)

        with self.assertRaises(ValidationError):
            DatasetLike.convert(42)

    def test_format(self):
        self.assertEqual(DatasetLike.format(None), '')

        with self.assertRaises(ValidationError):
            data = {'v1': [4, 5, 6], 'v2': [6, 7, 8]}
            DatasetLike.format(xr.Dataset(data_vars=data))


class DataFrameLikeTest(TestCase):
    def test_convert(self):
        self.assertEqual(DataFrameLike.convert(None), None)

        data = {'c1': [4, 5, 6], 'c2': [6, 7, 8]}
        xr_ds = xr.Dataset(data_vars=data)
        pd_ds = pd.DataFrame(data=data)
        gdf_ds = gpd.GeoDataFrame.from_features(read_test_features())
        proxy_gdf_ds = GeoDataFrame.from_features(read_test_features())
        self.assertIsInstance(DataFrameLike.convert(xr_ds), pd.DataFrame)
        self.assertIsInstance(DataFrameLike.convert(pd_ds), pd.DataFrame)
        self.assertIs(DataFrameLike.convert(pd_ds), pd_ds)
        self.assertIsInstance(DataFrameLike.convert(gdf_ds), gpd.GeoDataFrame)
        self.assertIs(DataFrameLike.convert(gdf_ds), gdf_ds)
        self.assertIsInstance(DataFrameLike.convert(proxy_gdf_ds), GeoDataFrame)
        self.assertIs(DataFrameLike.convert(proxy_gdf_ds), proxy_gdf_ds)

        with self.assertRaises(ValidationError):
            DataFrameLike.convert(42)

    def test_format(self):
        self.assertEqual(DataFrameLike.format(None), '')

        with self.assertRaises(ValidationError):
            data = {'c1': [4, 5, 6], 'c2': [6, 7, 8]}
            DataFrameLike.format(pd.DataFrame(data=data))


class TestGeoDataFrame(TestCase):

    def test_compatibility_with_geopandas(self):
        features = read_test_features()
        gdf = GeoDataFrame.from_features(features)
        self.assertIs(type(gdf), GeoDataFrame)
        self.assertIsInstance(gdf, GeoDataFrame)
        self.assertIsInstance(gdf, gpd.GeoDataFrame)
        self.assertIsInstance(gdf, pd.DataFrame)
        self.assertIs(gdf.features, features)
        self.assertIsInstance(gdf['A'], pd.Series)
        self.assertIsInstance(gdf.geometry, gpd.GeoSeries)

    def test_close(self):
        features = read_test_features()
        gdf = GeoDataFrame.from_features(features)
        self.assertIs(gdf.features, features)
        self.assertIsInstance(gdf.lazy_data_frame, gpd.GeoDataFrame)
        gdf.close()
        self.assertIsNone(gdf.features)
        self.assertIsNone(gdf.lazy_data_frame)


def read_test_features():
    import fiona
    import os
    return fiona.open(
        os.path.join(os.path.dirname(__file__), 'test_data', 'test.geojson')
    )



# The MIT License (MIT)
# Copyright (c) 2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

Helpers for data visualization operations (plot and animate).

Components
==========

"""
from esa_climate_toolbox.core.types import PolygonLike, ValidationError
from esa_climate_toolbox.core.opimpl import get_extents
from esa_climate_toolbox.util.im import ensure_cmaps_loaded


def handle_plot_polygon(region: PolygonLike.TYPE = None):
    """
    Return extents of the given PolygonLike.

    :param region: PolygonLike to introspect
    :return: extents
    """
    if region is None:
        return None

    extents, explicit_coords = get_extents(region)

    lon_min, lat_min, lon_max, lat_max = extents

    if not check_bounding_box(lat_min, lat_max, lon_min, lon_max):
        raise ValidationError('Provided plot extents do not form a valid bounding box '
                              'within [-180.0,+180.0,-90.0,+90.0]')
    return extents


def check_bounding_box(lat_min: float,
                       lat_max: float,
                       lon_min: float,
                       lon_max: float) -> bool:
    """
    Check if the provided [lat_min, lat_max, lon_min, lon_max] extents
    are sane.
    """
    if lat_min >= lat_max:
        return False

    if lon_min >= lon_max:
        return False

    if lat_min < -90.0:
        return False

    if lat_max > 90.0:
        return False

    if lon_min < -180.0:
        return False

    if lon_max > 180.0:
        return False

    return True


def in_notebook():
    """
    Return ``True`` if the module is running in IPython kernel,
    ``False`` if in IPython shell or other Python shell.
    """
    import sys
    ipykernel_in_sys_modules = 'ipykernel' in sys.modules
    # print('###########################################', ipykernel_in_sys_modules)
    return ipykernel_in_sys_modules


def get_var_data(var, indexers: dict, remaining_dims=None):
    """Select an arbitrary piece of an xarray dataset by using indexers."""
    if indexers:
        if remaining_dims:
            for dim in remaining_dims:
                if dim not in var.dims:
                    raise ValidationError(f'The specified dataset does not have a dimension called \'{dim}\'.')
                if dim in indexers:
                    raise ValidationError(f'Dimension \'{dim}\' is also specified as indexers. Please ensure that a '
                                          f'dimension is used exclusively either as indexers or as the selected '
                                          f'dimension.')
        for dim in indexers:
            if dim not in var.dims:
                raise ValidationError(f'The specified dataset does not have a dimension called \'{dim}\'.')
        var = var.sel(method='nearest', **indexers)

    if remaining_dims:
        isel_indexers = {dim_name: 0 for dim_name in var.dims if dim_name not in remaining_dims}
        var = var.isel(**isel_indexers)

    return var


def get_vars_data(ds, indexers: dict, remaining_dims=None):
    """Select an arbitrary piece of an xarray dataset by using indexers."""
    # to avoid the original dataset being affected (especially useful in unit tests)
    ds = ds.copy()

    if indexers:
        invalid_indexers = list(indexers)
        for var_name in ds:
            if ds[var_name].name in ds[var_name].dims:
                continue
            var_indexers = {}
            if remaining_dims:
                for dim in remaining_dims:
                    if dim not in ds[var_name].dims:
                        raise ValidationError(f'The specified dataset does not have a dimension called \'{dim}\'.')
                    if dim in indexers:
                        raise ValidationError(
                            f'Dimension \'{dim}\' is also specified as indexers. Please ensure that a '
                            f'dimension is used exclusively either as indexers or as the selected '
                            f'dimension.')
            for dim in ds[var_name].dims:
                if dim in indexers:
                    var_indexers[dim] = indexers[dim]
            for dim in invalid_indexers:
                if dim in ds[var_name].dims:
                    invalid_indexers.remove(dim)
            ds[var_name] = ds[var_name].sel(method='nearest', **var_indexers)

            if remaining_dims:
                isel_indexers = {dim_name: 0 for dim_name in ds[var_name].dims if dim_name not in remaining_dims}
                ds[var_name] = ds[var_name].isel(**isel_indexers)

        if len(invalid_indexers) > 0:
            raise ValidationError(f'There are dimensions specified in indexers but do not match dimensions in '
                                  f'any variables: {invalid_indexers}')

    return ds


# determine_cmap_params is adapted from Xarray through Seaborn:
# https://github.com/pydata/xarray/blob/master/xarray/plot/utils.py#L151
# https://github.com/mwaskom/seaborn/blob/v0.6/seaborn/matrix.py#L158
# Used under the terms of Seaborn's license:
# https://github.com/mwaskom/seaborn/blob/v0.8.1/LICENSE
#
# _determine_extend, _build_discrete_cmap, _color_palette, _is_scalar are
# adapted from Xarray and used under Xarray license:
# https://github.com/pydata/xarray/blob/v0.10.0/LICENSE

ROBUST_QUANTILE = 0.02


def determine_cmap_params(data_min, data_max, vmin=None, vmax=None, cmap=None,
                          center=None, robust=False, extend=None,
                          levels=None, filled=True, norm=None):
    """
    Use some heuristics to set good defaults for colorbar and range.

    :param plot_data: Data to use to determine colormap parameters
    :return: A dictionary containing calculated parameters
    """
    import matplotlib as mpl
    import numpy as np

    # Setting center=False prevents a divergent cmap
    possibly_divergent = center is not False

    # Set center to 0 so math below makes sense but remember its state
    center_is_none = False
    if center is None:
        center = 0
        center_is_none = True

    # Setting both vmin and vmax prevents a divergent cmap
    if (vmin is not None) and (vmax is not None):
        possibly_divergent = False

    # Setting vmin or vmax implies linspaced levels
    user_minmax = (vmin is not None) or (vmax is not None)

    # vlim might be computed below
    vlim = None

    if vmin is None:
        vmin = data_min
    elif possibly_divergent:
        vlim = abs(vmin - center)

    if vmax is None:
        vmax = data_max
    elif possibly_divergent:
        vlim = abs(vmax - center)

    if possibly_divergent:
        # kwargs not specific about divergent or not: infer defaults from data
        divergent = ((vmin < 0) and (vmax > 0)) or not center_is_none
    else:
        divergent = False

    # A divergent map should be symmetric around the center value
    if divergent:
        if vlim is None:
            vlim = max(abs(vmin - center), abs(vmax - center))
        vmin, vmax = -vlim, vlim

    # Now add in the centering value and set the limits
    vmin += center
    vmax += center

    # Choose default colormaps if not provided
    if cmap is None:
        if divergent:
            cmap = "RdBu_r"
        else:
            cmap = "viridis"

    # Handle discrete levels
    if levels is not None:
        if _is_scalar(levels):
            if user_minmax or levels == 1:
                levels = np.linspace(vmin, vmax, levels)
            else:
                # N in MaxNLocator refers to bins, not ticks
                ticker = mpl.ticker.MaxNLocator(levels - 1)
                levels = ticker.tick_values(vmin, vmax)
        vmin, vmax = levels[0], levels[-1]

    if extend is None:
        extend = _determine_extend(data_min, data_max, vmin, vmax)

    if levels is not None:
        cmap, norm = _build_discrete_cmap(cmap, levels, extend, filled)

    return dict(vmin=vmin, vmax=vmax, cmap=cmap, extend=extend,
                levels=levels, norm=norm)


def _determine_extend(data_min, data_max, vmin, vmax):
    extend_min = data_min < vmin
    extend_max = data_max > vmax
    if extend_min and extend_max:
        extend = 'both'
    elif extend_min:
        extend = 'min'
    elif extend_max:
        extend = 'max'
    else:
        extend = 'neither'
    return extend


def _build_discrete_cmap(cmap, levels, extend, filled):
    """Build a discrete colormap and normalization of the data."""
    import matplotlib as mpl

    if not filled:
        # non-filled contour plots
        extend = 'max'

    if extend == 'both':
        ext_n = 2
    elif extend in ['min', 'max']:
        ext_n = 1
    else:
        ext_n = 0

    n_colors = len(levels) + ext_n - 1
    pal = _color_palette(cmap, n_colors)

    new_cmap, cnorm = mpl.colors.from_levels_and_colors(
        levels, pal, extend=extend)
    # copy the old cmap name, for easier testing
    new_cmap.name = getattr(cmap, 'name', cmap)

    return new_cmap, cnorm


def _color_palette(cmap, n_colors):
    import matplotlib.pyplot as plt
    from matplotlib.colors import ListedColormap
    import numpy as np

    colors_i = np.linspace(0, 1., n_colors)
    if isinstance(cmap, (list, tuple)):
        # we have a list of colors
        cmap = ListedColormap(cmap, N=n_colors)
        pal = cmap(colors_i)
    elif isinstance(cmap, str):
        # we have some sort of named palette
        try:
            # is this a matplotlib cmap?
            ensure_cmaps_loaded()
            cmap = plt.get_cmap(cmap)
        except ValueError:
            # or maybe we just got a single color as a string
            cmap = ListedColormap([cmap], N=n_colors)
        pal = cmap(colors_i)
    else:
        # cmap better be a LinearSegmentedColormap (e.g. viridis)
        pal = cmap(colors_i)

    return pal


def _is_scalar(value):
    """Whether to treat a value as a scalar.
    Any non-iterable, string, or 0-D array
    """
    from collections.abc import Iterable

    return (getattr(value, 'ndim', None) == 0
            or isinstance(value, (str, bytes))
            or not isinstance(value, (Iterable,)))



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Tuple
import math

# max number of significant digits for a 64-bit float
_MAX_SIGNIFICANT_DIGITS_AFTER_DOT = 15

_MIN_EXP = -323
_MAX_EXP = 308


def sround(value: float, num_digits: int = 0, int_part=False) -> float:
    """
    Round *value* to significant number of digits *num_digits*.

    :param value: The value to round.
    :param num_digits: The number of digits after the first significant digit.
    :param int_part:
    :return:
    """
    num_digits_extra = _num_digits_extra(value, int_part=int_part)
    num_digits += num_digits_extra
    return round(value, ndigits=num_digits)


def sround_range(range_value: Tuple[float, float],
                 num_digits: int = 0,
                 int_part=False) -> Tuple[float, float]:
    value_1, value_2 = range_value
    num_digits_extra_1 = _num_digits_extra(value_1, int_part=int_part)
    num_digits_extra_2 = _num_digits_extra(value_2, int_part=int_part)
    num_digits += min(num_digits_extra_1, num_digits_extra_2)
    return round(value_1, ndigits=num_digits), \
           round(value_2, ndigits=num_digits)


def _num_digits_extra(value: float, int_part: bool) -> int:
    num_digits = -int(math.floor(_limited_log10(value)))
    if num_digits < 0 and not int_part:
        return 0
    return num_digits


def _limited_log10(value: float) -> float:
    if value > 0.0:
        exp = math.log10(value)
    elif value < 0.0:
        exp = math.log10(-value)
    else:
        return _MIN_EXP

    if exp < _MIN_EXP:
        return _MIN_EXP
    if exp > _MAX_EXP:
        return _MAX_EXP

    return exp



"""
Tests for aggregation operations
"""
from unittest import TestCase

import numpy as np
import pandas as pd
import xarray as xr

from esa_climate_toolbox.core.op import OP_REGISTRY
from esa_climate_toolbox.ops import adjust_temporal_attrs
from esa_climate_toolbox.ops import climatology, temporal_aggregation, reduce
from esa_climate_toolbox.util.misc import object_to_qualified_name
from esa_climate_toolbox.util.monitor import ConsoleMonitor


class TestClimatology(TestCase):
    """
    Test long term averaging
    """

    def test_nominal(self):
        """
        Test nominal execution
        """
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 24])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 24])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', freq='MS', periods=24)})

        ds = adjust_temporal_attrs(ds)

        # Test monitor
        m = ConsoleMonitor()
        actual = climatology(ds, monitor=m)
        self.assertEqual(m._percentage, 100)

        # Test CF attributes
        self.assertEqual(actual.dims, {'time': 12,
                                       'nv': 2,
                                       'lat': 45,
                                       'lon': 90})
        self.assertEqual(actual.time.attrs['climatology'],
                         'climatology_bounds')

        # Test variable selection
        actual = climatology(ds, var='first')
        with self.assertRaises(KeyError):
            actual['second']

    # @unittest.skip("Daily aggregation does do weird things. Skipping for the moment")
    def test_daily(self):
        """
        Test creating a daily LTA dataset
        """
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 730])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 730])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2001-01-01', '2002-12-31')})
        ds = adjust_temporal_attrs(ds)
        actual = climatology(ds)

        self.assertEqual(actual.dims, {'dayofyear': 365,
                                       'lat': 45,
                                       'lon': 90})

    def test_general(self):
        """
        Test creating a 'general' LTA dataset
        """
        # Test seasonal
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 5])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 5])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('1999-12-01', freq='QS-DEC', periods=5)})
        ds = adjust_temporal_attrs(ds)
        actual = climatology(ds)
        self.assertEqual(actual['first'].attrs['cell_methods'],
                         'time: mean over years')
        self.assertEqual(actual.dims, {'time': 4,
                                       'nv': 2,
                                       'lat': 45,
                                       'lon': 90})
        self.assertEqual(actual.time.attrs['climatology'],
                         'climatology_bounds')

        # Test irregular seasons
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 6])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 6])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', freq='5M', periods=6)})
        ds = adjust_temporal_attrs(ds)
        with self.assertRaises(ValueError) as err:
            climatology(ds)
        self.assertIn('inconsistent seasons', str(err.exception))

        # Test 8D
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 137])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 137])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', '2002-12-31', freq='8D')})
        ds = adjust_temporal_attrs(ds)
        with self.assertRaises(ValueError) as err:
            climatology(ds)
        self.assertIn('inconsistent seasons', str(err.exception))

    def test_registered(self):
        """
        Test registered operation execution
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(climatology))
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 24])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 24])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', freq='MS', periods=24)})

        ds = adjust_temporal_attrs(ds)

        reg_op(ds=ds)

    def test_validation(self):
        """
        Test input validation
        """
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 24])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})

        ds = adjust_temporal_attrs(ds)

        with self.assertRaises(ValueError) as err:
            climatology(ds)
        self.assertIn('normalize', str(err.exception))


class TestTemporalAggregation(TestCase):
    """
    Test temporal aggregation
    """

    def test_nominal(self):
        """
        Test nominal exeuction
        """
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 366])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 366])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', '2000-12-31')})
        ds = adjust_temporal_attrs(ds)

        ex = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', freq='MS', periods=12)})
        ex.first.attrs['cell_methods'] = 'time: mean within years'
        ex.second.attrs['cell_methods'] = 'time: mean within years'

        m = ConsoleMonitor()
        actual = temporal_aggregation(ds, monitor=m)

        self.assertTrue(actual.broadcast_equals(ex))

    def test_seasonal(self):
        """
        Test aggregation to a seasonal dataset
        """
        # daily -> seasonal
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 366])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 366])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', '2000-12-31')})
        ds = adjust_temporal_attrs(ds)

        ex = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 5])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 5])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('1999-12-01', freq='QS-DEC', periods=5)})
        m = ConsoleMonitor()
        actual = temporal_aggregation(ds, period='QS-DEC', monitor=m)
        self.assertTrue(actual.broadcast_equals(ex))

        # TODO: uncomment once we have a fix for
        #  esa_climate_toolbox.core.types.ValidationError:
        #  Units 'M', 'Y', and 'y' are no longer supported,
        #  as they do not represent unambiguous timedelta values durations.

        # # monthly -> seasonal
        # ds = xr.Dataset({
        #     'first': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
        #     'second': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
        #     'lat': np.linspace(-88, 88, 45),
        #     'lon': np.linspace(-178, 178, 90),
        #     'time': pd.date_range('2000-01-01', freq='MS', periods=12)})
        # ds = adjust_temporal_attrs(ds)
        # actual = temporal_aggregation(ds, output_resolution='season', monitor=m)
        # self.assertTrue(actual.broadcast_equals(ex))

    def test_custom(self):
        """
        Test aggergating to custom temporal resolutions
        """
        # daily -> 8 days
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 366])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 366])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', '2000-12-31')})
        ds = adjust_temporal_attrs(ds)

        ex = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 46])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 46])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', '2000-12-31', freq='8D')})
        ex.first.attrs['cell_methods'] = 'time: mean within years'
        ex.second.attrs['cell_methods'] = 'time: mean within years'
        m = ConsoleMonitor()
        actual = temporal_aggregation(ds, period='8D', monitor=m)
        self.assertTrue(actual.broadcast_equals(ex))

        # daily -> weekly
        ex = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 53])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 53])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', '2000-12-31', freq='W')})
        ex.first.attrs['cell_methods'] = 'time: mean within years'
        ex.second.attrs['cell_methods'] = 'time: mean within years'
        actual = temporal_aggregation(ds, period='W', monitor=m)
        self.assertTrue(actual.broadcast_equals(ex))

        # TODO: uncomment once we have a fix for
        #  esa_climate_toolbox.core.types.ValidationError:
        #  Units 'M', 'Y', and 'y' are no longer supported,
        #  as they do not represent unambiguous timedelta values durations.

        # # monthly -> 4 month seasons
        # ds = xr.Dataset({
        #     'first': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
        #     'second': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
        #     'lat': np.linspace(-88, 88, 45),
        #     'lon': np.linspace(-178, 178, 90),
        #     'time': pd.date_range('2000-01-01', freq='MS', periods=12)})
        # ds = adjust_temporal_attrs(ds)
        # ex = xr.Dataset({
        #     'first': (['lat', 'lon', 'time'], np.ones([45, 90, 4])),
        #     'second': (['lat', 'lon', 'time'], np.ones([45, 90, 4])),
        #     'lat': np.linspace(-88, 88, 45),
        #     'lon': np.linspace(-178, 178, 90),
        #     'time': pd.date_range('2000-01-01', freq='4M', periods=4)})
        # ex.first.attrs['cell_methods'] = 'time: mean within years'
        # ex.second.attrs['cell_methods'] = 'time: mean within years'
        # actual = temporal_aggregation(ds, custom_resolution='4M', monitor=m)
        # self.assertTrue(actual.broadcast_equals(ex))

    def test_8days(self):
        """
        Test nominal execution with a 8 Days frequency input dataset
        """
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 46])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 46])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', '2000-12-31', freq='8D')})
        ds = adjust_temporal_attrs(ds)

        ex = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', freq='MS', periods=12)})
        ex.first.attrs['cell_methods'] = 'time: mean within years'
        ex.second.attrs['cell_methods'] = 'time: mean within years'

        m = ConsoleMonitor()
        actual = temporal_aggregation(ds, monitor=m)

        self.assertTrue(actual.broadcast_equals(ex))

    def test_registered(self):
        """
        Test registered operation execution
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(temporal_aggregation))
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 366])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 366])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', '2000-12-31')})

        ex = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 12])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', freq='MS', periods=12)})

        actual = reg_op(ds=ds)
        self.assertTrue(actual.broadcast_equals(ex))

    def test_validation(self):
        """
        Test input validation
        """
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 24])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90)})
        ds = adjust_temporal_attrs(ds)

        with self.assertRaises(ValueError) as err:
            temporal_aggregation(ds)
        self.assertIn('normalize', str(err.exception))

        # invalid custom resolution
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 366])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 366])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', '2000-12-31')})
        ds = adjust_temporal_attrs(ds)

        with self.assertRaises(ValueError) as err:
            temporal_aggregation(ds, period='SDFG')
        self.assertEqual('Invalid freq value "SDFG". '
                         'Please check operation documentation.', str(err.exception))


class TestReduce(TestCase):
    """
    Test reduce() operation
    """

    def test_nominal(self):
        """
        Test nominal execution
        """
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.ones([45, 90, 366])),
            'second': (['lat', 'lon', 'time'], np.ones([45, 90, 366])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', '2000-12-31')})
        ds = adjust_temporal_attrs(ds)

        ex = xr.Dataset({
            'first': (['lat', 'lon'], np.ones([45, 90])),
            'second': (['lat', 'lon'], np.ones([45, 90])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', '2000-12-31')})

        m = ConsoleMonitor()
        actual = reduce(ds, monitor=m)

        self.assertTrue(actual.broadcast_equals(ex))

        actual = reduce(ds, dim=['lat', 'time', 'there_is_no_spoon'], monitor=m)
        ex = xr.Dataset({
            'first': (['lon'], np.ones([90])),
            'second': (['lon'], np.ones([90])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': pd.date_range('2000-01-01', '2000-12-31')})

        self.assertTrue(actual.broadcast_equals(ex))



import os
import shutil
import unittest

from unittest import skipIf

from xcube.core.new import new_cube
from xcube.core.store import DataStore

from esa_climate_toolbox.constants import ECT_STORE_ID
from esa_climate_toolbox.constants import ECT_KC_STORE_ID
from esa_climate_toolbox.constants import ECT_ZARR_STORE_ID

from esa_climate_toolbox.core.ds import add_local_store
from esa_climate_toolbox.core.ds import get_store
from esa_climate_toolbox.core.ds import list_datasets
from esa_climate_toolbox.core.ds import list_ecvs
from esa_climate_toolbox.core.ds import list_ecv_datasets
from esa_climate_toolbox.core.ds import list_stores
from esa_climate_toolbox.core.ds import remove_store
from esa_climate_toolbox.core.ds import write_data
from esa_climate_toolbox.core.types import ValidationError


class DsTest(unittest.TestCase):

    def test_list_stores(self):
        provided_stores = list_stores()
        self.assertIn(ECT_STORE_ID, provided_stores)
        self.assertIn(ECT_KC_STORE_ID, provided_stores)
        self.assertIn(ECT_ZARR_STORE_ID, provided_stores)

    @skipIf(os.environ.get('ECT_DISABLE_WEB_TESTS', '1') == '1',
            'ECT_DISABLE_WEB_TESTS = 1')
    def test_list_ecvs(self):
        ecvs = list_ecvs()
        expected_ecvs = [
            'SEAICE', 'ICESHEETS', 'BIOMASS', 'CLOUD', 'FIRE', 'SEALEVEL',
            'WATERVAPOUR', 'LC', 'OZONE', 'LST', 'OC', 'LAKES', 'SNOW', 'SST',
            'GHG', 'SEASTATE', 'AEROSOL', 'SEASURFACESALINITY', 'PERMAFROST',
            'SOILMOISTURE'
        ]
        for expected_ecv in expected_ecvs:
            self.assertIn(expected_ecv, ecvs)

    @skipIf(os.environ.get('ECT_DISABLE_WEB_TESTS', '1') == '1',
            'ECT_DISABLE_WEB_TESTS = 1')
    def test_list_ecv_datasets_fail(self):
        with self.assertRaises(ValueError) as ve:
            list_ecv_datasets('Wind')
        expected = \
            '"Wind" is not an Essential Climate Variable provided by ' \
            'the ESA Climate Toolbox. Please choose one of the following:'
        self.assertEqual(expected, str(ve.exception)[:len(expected)])

    @skipIf(os.environ.get('ECT_DISABLE_WEB_TESTS', '1') == '1',
            'ECT_DISABLE_WEB_TESTS = 1')
    def test_list_ecv_datasets_lc(self):
        lakes_datasets = list_ecv_datasets('LC')
        expected_zarr_file = (
            'ESACCI-LC-L4-LCCS-Map-300m-P1Y-1992-2015-v2.0.7b.zarr',
            ECT_ZARR_STORE_ID
        )
        self.assertIn(expected_zarr_file, lakes_datasets)

    def test_get_store(self):
        store = get_store(store_id=ECT_STORE_ID)
        self.assertIsNotNone(store)
        self.assertIsInstance(store, DataStore)

    @skipIf(os.environ.get('ECT_DISABLE_WEB_TESTS', '1') == '1',
            'ECT_DISABLE_WEB_TESTS = 1')
    def test_list_datasets(self):
        datasets = list_datasets(store_id=ECT_KC_STORE_ID)
        self.assertTrue(len(datasets) > 1)
        self.assertTrue(
            'ESACCI-LC-L4-PFT-Map-300m-P1Y-1992-2020-v2.0.8-kr1.1'
            in datasets
        )

        datasets = list_datasets(store_id=ECT_ZARR_STORE_ID)
        self.assertTrue(len(datasets) > 1)
        self.assertTrue(
            'ESACCI-LC-L4-LCCS-Map-300m-P1Y-1992-2015-v2.0.7b.zarr'
            in datasets
        )


class DsStoresTest(unittest.TestCase):

    def setUp(self) -> None:
        self._root = os.path.join(os.path.dirname(__file__), 'test_data')

    def test_add_local_store_id_taken(self):
        with self.assertRaises(ValueError) as ve:
            add_local_store(root=self._root, store_id=ECT_STORE_ID)
        self.assertEqual(
            f'There is already a store with the id "{ECT_STORE_ID}" '
            f'registered. Please choose another id.',
            str(ve.exception)
        )

    def test_remove_store_essential(self):
        with self.assertRaises(ValueError) as ve:
            remove_store(ECT_STORE_ID, persist=False)
        self.assertEqual(
            f'Cannot remove essential store "{ECT_STORE_ID}".',
            str(ve.exception)
        )

    def test_remove_store_missing(self):
        with self.assertRaises(ValueError) as ve:
            remove_store('vtmcj', persist=False)
        self.assertEqual(
            'No store named "vtmcj" found.',
            str(ve.exception)
        )

    def test_full_workflow_local(self):
        local_store_id = add_local_store(
            root=self._root, persist=False
        )
        self.assertTrue(local_store_id.startswith('local_'))
        self.assertIn(local_store_id, list_stores())
        self.assertNotIn(
            'small/'
            'ESACCI-SOILMOISTURE-L3S-SSMV-COMBINED-20000101000000-fv02.2.nc',
            list_datasets(local_store_id)
        )

        other_local_store_id = add_local_store(
            root=self._root, max_depth=2, persist=False
        )

        self.assertTrue(other_local_store_id.startswith('local_'))
        self.assertIn(other_local_store_id, list_stores())
        self.assertIn(
            'small/'
            'ESACCI-SOILMOISTURE-L3S-SSMV-COMBINED-20000101000000-fv02.2.nc',
            list_datasets(other_local_store_id)
        )
        self.assertIn(
            'small/'
            'ESACCI-SOILMOISTURE-L3S-SSMV-COMBINED-20000101000000-fv02.2.nc',
            list_datasets(other_local_store_id, data_type='dataset')
        )
        self.assertNotIn(
            'small/'
            'ESACCI-SOILMOISTURE-L3S-SSMV-COMBINED-20000101000000-fv02.2.nc',
            list_datasets(other_local_store_id, data_type='geodataframe')
        )

        remove_store(local_store_id, persist=False)
        self.assertFalse(local_store_id in list_stores())
        remove_store(other_local_store_id, persist=False)
        self.assertFalse(other_local_store_id in list_stores())

    def test_full_workflow_local_zarr_dataset_paths(self):
        local_zarr_store_id = add_local_store(
            root=self._root, max_depth=2, includes='*.zarr',
            persist=False
        )
        self.assertTrue(local_zarr_store_id.startswith('local_'))
        self.assertIn(local_zarr_store_id, list_stores())
        self.assertEqual([], list_datasets(local_zarr_store_id))
        remove_store(local_zarr_store_id, persist=False)
        self.assertFalse(local_zarr_store_id in list_stores())

    def test_full_workflow_local_nc_dataset_paths(self):
        local_nc_store_id = add_local_store(
            root=self._root, max_depth=2, includes='*.nc',
            persist=False
        )
        self.assertTrue(local_nc_store_id.startswith('local_'))
        self.assertIn(local_nc_store_id, list_stores())
        self.assertIn(
            'small/'
            'ESACCI-SOILMOISTURE-L3S-SSMV-COMBINED-20000101000000-fv02.2.nc',
            list_datasets(local_nc_store_id)
        )
        remove_store(local_nc_store_id, persist=False)
        self.assertFalse(local_nc_store_id in list_stores())


class WriteTest(unittest.TestCase):

    def setUp(self) -> None:
        self.cube = new_cube()
        self.path = os.path.join(os.getcwd(), 'test_path')
        self.local_store_id = add_local_store(root=self.path, persist=False)
        os.makedirs(self.path, exist_ok=True)

    def tearDown(self) -> None:
        remove_store(store_id=self.local_store_id)
        shutil.rmtree(self.path)

    def test_write_no_output_store_set(self):
        with self.assertRaises(ValueError) as ve:
            write_data(self.cube)
        self.assertEqual(
            'No default output store set. Must specify target store id.',
            str(ve.exception)
        )

    def test_write_invalid_output_store(self):
        with self.assertRaises(TypeError) as te:
            write_data(self.cube, store_id=ECT_STORE_ID)
        self.assertEqual(
            "value must be an instance of "
            "<class 'xcube.core.store.store.MutableDataStore'>, "
            "was <class 'esa_climate_toolbox.ds.dataaccess.CciCdcDataStore'>",
            str(te.exception)
        )

    def test_write_wrong_format(self):
        with self.assertRaises(ValidationError) as ve:
            write_data(
                self.cube, store_id=self.local_store_id, format_id='geojson'
            )
        self.assertEqual(
            f'Format "geojson" is not supported by data store '
            f'"{self.local_store_id}". Must be one of the following: '
            f'netcdf, zarr, levels, geotiff',
            str(ve.exception)
        )

    def test_write_nc(self):
        data_id = write_data(
            self.cube, store_id=self.local_store_id, format_id='netcdf'
        )
        self.assertTrue(data_id.endswith('.nc'))
        self.assertIn(data_id, list_datasets(self.local_store_id))

    def test_write_zarr(self):
        data_id = write_data(
            self.cube, store_id=self.local_store_id
        )
        self.assertTrue(data_id.endswith('.zarr'))
        self.assertIn(data_id, list_datasets(self.local_store_id))






"""
Test for the normalisation operation
"""

import calendar
from datetime import datetime
from unittest import TestCase

import cftime
import numpy as np
import pandas as pd
import xarray as xr
from jdcal import gcal2jd
from numpy.testing import assert_array_almost_equal

from esa_climate_toolbox.core.op import OP_REGISTRY
from esa_climate_toolbox.ops.normalize import adjust_spatial_attrs
from esa_climate_toolbox.ops.normalize import adjust_temporal_attrs
from esa_climate_toolbox.ops.normalize import normalize
from esa_climate_toolbox.util.misc import object_to_qualified_name

from xcube.core.normalize import normalize_coord_vars
from xcube.core.normalize import normalize_missing_time


# noinspection PyPep8Naming
def assertDatasetEqual(expected, actual):
    # this method is functionally equivalent to
    # `assert expected == actual`, but it
    # checks each aspect of equality separately for easier debugging
    assert expected.equals(actual), (expected, actual)


class TestNormalize(TestCase):

    def test_normalize_lon_lat_2d(self):
        """
        Test nominal execution
        """
        dims = ('time', 'y', 'x')
        attributes = {'valid_min': 0., 'valid_max': 1.}

        t_size = 2
        y_size = 3
        x_size = 4

        a_data = np.random.random_sample((t_size, y_size, x_size))
        b_data = np.random.random_sample((t_size, y_size, x_size))
        time_data = [1, 2]
        lat_data = [[10., 10., 10., 10.],
                    [20., 20., 20., 20.],
                    [30., 30., 30., 30.]]
        lon_data = [[-10., 0., 10., 20.],
                    [-10., 0., 10., 20.],
                    [-10., 0., 10., 20.]]
        dataset = xr.Dataset({'a': (dims, a_data, attributes),
                              'b': (dims, b_data, attributes)
                              },
                             {'time': (('time',), time_data),
                              'lat': (('y', 'x'), lat_data),
                              'lon': (('y', 'x'), lon_data)
                              },
                             {'geospatial_lon_min': -15.,
                              'geospatial_lon_max': 25.,
                              'geospatial_lat_min': 5.,
                              'geospatial_lat_max': 35.
                              }
                             )

        new_dims = ('time', 'lat', 'lon')
        expected = xr.Dataset({'a': (new_dims, a_data, attributes),
                               'b': (new_dims, b_data, attributes)},
                              {'time': (('time',), time_data),
                               'lat': (('lat',), [10., 20., 30.]),
                               'lon': (('lon',), [-10., 0., 10., 20.]),
                               },
                              {'geospatial_lon_min': -15.,
                               'geospatial_lon_max': 25.,
                               'geospatial_lat_min': 5.,
                               'geospatial_lat_max': 35.})

        actual = normalize(dataset)
        xr.testing.assert_equal(actual, expected)

    def test_normalize_lon_lat(self):
        """
        Test nominal execution
        """
        dataset = xr.Dataset({'first': (['latitude',
                                         'longitude'], [[1, 2, 3],
                                                        [2, 3, 4]])})
        expected = xr.Dataset({'first': (['lat', 'lon'], [[1, 2, 3],
                                                          [2, 3, 4]])})
        actual = normalize(dataset)
        assertDatasetEqual(actual, expected)

        dataset = xr.Dataset({'first': (['lat', 'long'], [[1, 2, 3],
                                                          [2, 3, 4]])})
        expected = xr.Dataset({'first': (['lat', 'lon'], [[1, 2, 3],
                                                          [2, 3, 4]])})
        actual = normalize(dataset)
        assertDatasetEqual(actual, expected)

        dataset = xr.Dataset({'first': (['latitude',
                                         'spacetime'], [[1, 2, 3],
                                                        [2, 3, 4]])})
        expected = xr.Dataset({'first': (['lat', 'spacetime'], [[1, 2, 3],
                                                                [2, 3, 4]])})
        actual = normalize(dataset)
        assertDatasetEqual(actual, expected)

        dataset = xr.Dataset({'first': (['zef', 'spacetime'], [[1, 2, 3],
                                                               [2, 3, 4]])})
        expected = xr.Dataset({'first': (['zef', 'spacetime'], [[1, 2, 3],
                                                                [2, 3, 4]])})
        actual = normalize(dataset)
        assertDatasetEqual(actual, expected)

    def test_normalize_inverted_lat(self):
        first = np.zeros([3, 45, 90])
        first[0, :, :] = np.eye(45, 90)
        ds = xr.Dataset({
            'first': (['time', 'lat', 'lon'], first),
            'second': (['time', 'lat', 'lon'], np.zeros([3, 45, 90])),
            'lat': np.linspace(88, -88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': [datetime(2000, x, 1) for x in range(1, 4)]
        }).chunk(chunks={'time': 1})

        first = np.zeros([3, 45, 90])
        first[0, :, :] = np.flip(np.eye(45, 90), axis=0)
        expected = xr.Dataset({
            'first': (['time', 'lat', 'lon'], first),
            'second': (['time', 'lat', 'lon'], np.zeros([3, 45, 90])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': [datetime(2000, x, 1) for x in range(1, 4)]
        }).chunk(chunks={'time': 1})

        actual = normalize(ds)
        xr.testing.assert_equal(actual, expected)

    def test_normalize_with_missing_time_dim(self):
        ds = xr.Dataset({'first': (['lat', 'lon'], np.zeros([90, 180])),
                         'second': (['lat', 'lon'], np.zeros([90, 180]))},
                        coords={'lat': np.linspace(-89.5, 89.5, 90),
                                'lon': np.linspace(-179.5, 179.5, 180)},
                        attrs={'time_coverage_start': '20120101',
                               'time_coverage_end': '20121231'})
        norm_ds = normalize(ds)
        self.assertIsNot(norm_ds, ds)
        self.assertEqual(len(norm_ds.coords), 4)
        self.assertIn('lon', norm_ds.coords)
        self.assertIn('lat', norm_ds.coords)
        self.assertIn('time', norm_ds.coords)
        self.assertIn('time_bnds', norm_ds.coords)

        self.assertEqual(norm_ds.first.shape, (1, 90, 180))
        self.assertEqual(norm_ds.second.shape, (1, 90, 180))
        self.assertEqual(norm_ds.coords['time'][0],
                         xr.DataArray(pd.to_datetime('2012-07-01T12:00:00')))
        self.assertEqual(norm_ds.coords['time_bnds'][0][0],
                         xr.DataArray(pd.to_datetime('2012-01-01')))
        self.assertEqual(norm_ds.coords['time_bnds'][0][1],
                         xr.DataArray(pd.to_datetime('2012-12-31')))

    def test_normalize_with_time_called_t(self):
        ds = xr.Dataset(
            {'first': (['time', 'lat', 'lon'], np.zeros([4, 90, 180])),
             'second': (['time', 'lat', 'lon'], np.zeros([4, 90, 180])),
             't': ('time', np.array(['2005-07-02T00:00:00.000000000',
                                     '2006-07-02T12:00:00.000000000',
                                     '2007-07-03T00:00:00.000000000',
                                     '2008-07-02T00:00:00.000000000'],
                                    dtype='datetime64[ns]'))
             },
            coords={'lat': np.linspace(-89.5, 89.5, 90),
                    'lon': np.linspace(-179.5, 179.5, 180)},
            attrs={'time_coverage_start': '2005-01-17',
                   'time_coverage_end': '2008-08-17'}
        )
        norm_ds = normalize(ds)
        self.assertIsNot(norm_ds, ds)
        self.assertEqual(len(norm_ds.coords), 3)
        self.assertIn('lon', norm_ds.coords)
        self.assertIn('lat', norm_ds.coords)
        self.assertIn('time', norm_ds.coords)

        self.assertEqual(norm_ds.first.shape, (4, 90, 180))
        self.assertEqual(norm_ds.second.shape, (4, 90, 180))
        self.assertEqual(norm_ds.coords['time'][0],
                         xr.DataArray(pd.to_datetime('2005-07-02T00:00')))

    def test_normalize_julian_day(self):
        """
        Test Julian Day -> Datetime conversion
        """
        tuples = [gcal2jd(2000, x, 1) for x in range(1, 13)]

        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'second': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': [x[0] + x[1] for x in tuples]})
        ds.time.attrs['long_name'] = 'time in julian days'

        expected = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.zeros([12, 45, 90])),
            'second': (['time', 'lat', 'lon'], np.zeros([12, 45, 90])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': [datetime(2000, x, 1) for x in range(1, 13)]})
        expected.time.attrs['long_name'] = 'time'

        actual = normalize(ds)

        assertDatasetEqual(actual, expected)

    def test_registered(self):
        """
        Test as a registered operation
        """
        reg_op = OP_REGISTRY.get_op(object_to_qualified_name(normalize))
        dataset = xr.Dataset({'first': (['latitude',
                                         'longitude'], [[1, 2, 3],
                                                        [2, 3, 4]])})
        expected = xr.Dataset({'first': (['lat', 'lon'], [[1, 2, 3],
                                                          [2, 3, 4]])})
        actual = reg_op(ds=dataset)
        assertDatasetEqual(actual, expected)


class TestAdjustSpatial(TestCase):
    def test_nominal(self):
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'second': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': [datetime(2000, x, 1) for x in range(1, 13)]})

        ds.lon.attrs['units'] = 'degrees_east'
        ds.lat.attrs['units'] = 'degrees_north'

        ds1 = adjust_spatial_attrs(ds)

        # Make sure original dataset is not altered
        with self.assertRaises(KeyError):
            # noinspection PyStatementEffect
            ds.attrs['geospatial_lat_min']

        # Make sure expected values are in the new dataset
        self.assertEqual(ds1.attrs['geospatial_lat_min'], -90)
        self.assertEqual(ds1.attrs['geospatial_lat_max'], 90)
        self.assertEqual(ds1.attrs['geospatial_lat_units'], 'degrees_north')
        self.assertEqual(ds1.attrs['geospatial_lat_resolution'], 4)
        self.assertEqual(ds1.attrs['geospatial_lon_min'], -180)
        self.assertEqual(ds1.attrs['geospatial_lon_max'], 180)
        self.assertEqual(ds1.attrs['geospatial_lon_units'], 'degrees_east')
        self.assertEqual(ds1.attrs['geospatial_lon_resolution'], 4)
        self.assertEqual(ds1.attrs['geospatial_bounds'],
                         'POLYGON((-180.0 -90.0, -180.0 90.0, 180.0 90.0,'
                         ' 180.0 -90.0, -180.0 -90.0))')

        # Test existing attributes update
        lon_min, lat_min, lon_max, lat_max = -20, -40, 60, 40
        indexers = {'lon': slice(lon_min, lon_max),
                    'lat': slice(lat_min, lat_max)}
        ds2 = ds1.sel(**indexers)
        ds2 = adjust_spatial_attrs(ds2)

        self.assertEqual(ds2.attrs['geospatial_lat_min'], -42)
        self.assertEqual(ds2.attrs['geospatial_lat_max'], 42)
        self.assertEqual(ds2.attrs['geospatial_lat_units'], 'degrees_north')
        self.assertEqual(ds2.attrs['geospatial_lat_resolution'], 4)
        self.assertEqual(ds2.attrs['geospatial_lon_min'], -20)
        self.assertEqual(ds2.attrs['geospatial_lon_max'], 60)
        self.assertEqual(ds2.attrs['geospatial_lon_units'], 'degrees_east')
        self.assertEqual(ds2.attrs['geospatial_lon_resolution'], 4)
        self.assertEqual(ds2.attrs['geospatial_bounds'],
                         'POLYGON((-20.0 -42.0, -20.0 42.0, 60.0 42.0, 60.0'
                         ' -42.0, -20.0 -42.0))')

    def test_nominal_inverted(self):
        # Inverted lat
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'second': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'lat': np.linspace(88, -88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': [datetime(2000, x, 1) for x in range(1, 13)]})

        ds.lon.attrs['units'] = 'degrees_east'
        ds.lat.attrs['units'] = 'degrees_north'

        ds1 = adjust_spatial_attrs(ds)

        # Make sure original dataset is not altered
        with self.assertRaises(KeyError):
            # noinspection PyStatementEffect
            ds.attrs['geospatial_lat_min']

        # Make sure expected values are in the new dataset
        self.assertEqual(ds1.attrs['geospatial_lat_min'], -90)
        self.assertEqual(ds1.attrs['geospatial_lat_max'], 90)
        self.assertEqual(ds1.attrs['geospatial_lat_units'], 'degrees_north')
        self.assertEqual(ds1.attrs['geospatial_lat_resolution'], 4)
        self.assertEqual(ds1.attrs['geospatial_lon_min'], -180)
        self.assertEqual(ds1.attrs['geospatial_lon_max'], 180)
        self.assertEqual(ds1.attrs['geospatial_lon_units'], 'degrees_east')
        self.assertEqual(ds1.attrs['geospatial_lon_resolution'], 4)
        self.assertEqual(ds1.attrs['geospatial_bounds'],
                         'POLYGON((-180.0 -90.0, -180.0 90.0, 180.0 90.0,'
                         ' 180.0 -90.0, -180.0 -90.0))')

        # Test existing attributes update
        lon_min, lat_min, lon_max, lat_max = -20, -40, 60, 40
        indexers = {'lon': slice(lon_min, lon_max),
                    'lat': slice(lat_max, lat_min)}
        ds2 = ds1.sel(**indexers)
        ds2 = adjust_spatial_attrs(ds2)

        self.assertEqual(ds2.attrs['geospatial_lat_min'], -42)
        self.assertEqual(ds2.attrs['geospatial_lat_max'], 42)
        self.assertEqual(ds2.attrs['geospatial_lat_units'], 'degrees_north')
        self.assertEqual(ds2.attrs['geospatial_lat_resolution'], 4)
        self.assertEqual(ds2.attrs['geospatial_lon_min'], -20)
        self.assertEqual(ds2.attrs['geospatial_lon_max'], 60)
        self.assertEqual(ds2.attrs['geospatial_lon_units'], 'degrees_east')
        self.assertEqual(ds2.attrs['geospatial_lon_resolution'], 4)
        self.assertEqual(ds2.attrs['geospatial_bounds'],
                         'POLYGON((-20.0 -42.0, -20.0 42.0, 60.0 42.0, 60.0'
                         ' -42.0, -20.0 -42.0))')

    def test_bnds(self):
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'second': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': [datetime(2000, x, 1) for x in range(1, 13)]})

        ds.lon.attrs['units'] = 'degrees_east'
        ds.lat.attrs['units'] = 'degrees_north'

        lat_bnds = np.empty([len(ds.lat), 2])
        lon_bnds = np.empty([len(ds.lon), 2])
        ds['nv'] = [0, 1]

        lat_bnds[:, 0] = ds.lat.values - 2
        lat_bnds[:, 1] = ds.lat.values + 2
        lon_bnds[:, 0] = ds.lon.values - 2
        lon_bnds[:, 1] = ds.lon.values + 2

        ds['lat_bnds'] = (['lat', 'nv'], lat_bnds)
        ds['lon_bnds'] = (['lon', 'nv'], lon_bnds)

        ds.lat.attrs['bounds'] = 'lat_bnds'
        ds.lon.attrs['bounds'] = 'lon_bnds'

        ds1 = adjust_spatial_attrs(ds)

        # Make sure original dataset is not altered
        with self.assertRaises(KeyError):
            # noinspection PyStatementEffect
            ds.attrs['geospatial_lat_min']

        # Make sure expected values are in the new dataset
        self.assertEqual(ds1.attrs['geospatial_lat_min'], -90)
        self.assertEqual(ds1.attrs['geospatial_lat_max'], 90)
        self.assertEqual(ds1.attrs['geospatial_lat_units'], 'degrees_north')
        self.assertEqual(ds1.attrs['geospatial_lat_resolution'], 4)
        self.assertEqual(ds1.attrs['geospatial_lon_min'], -180)
        self.assertEqual(ds1.attrs['geospatial_lon_max'], 180)
        self.assertEqual(ds1.attrs['geospatial_lon_units'], 'degrees_east')
        self.assertEqual(ds1.attrs['geospatial_lon_resolution'], 4)
        self.assertEqual(ds1.attrs['geospatial_bounds'],
                         'POLYGON((-180.0 -90.0, -180.0 90.0, 180.0 90.0,'
                         ' 180.0 -90.0, -180.0 -90.0))')

        # Test existing attributes update
        lon_min, lat_min, lon_max, lat_max = -20, -40, 60, 40
        indexers = {'lon': slice(lon_min, lon_max),
                    'lat': slice(lat_min, lat_max)}
        ds2 = ds1.sel(**indexers)
        ds2 = adjust_spatial_attrs(ds2)

        self.assertEqual(ds2.attrs['geospatial_lat_min'], -42)
        self.assertEqual(ds2.attrs['geospatial_lat_max'], 42)
        self.assertEqual(ds2.attrs['geospatial_lat_units'], 'degrees_north')
        self.assertEqual(ds2.attrs['geospatial_lat_resolution'], 4)
        self.assertEqual(ds2.attrs['geospatial_lon_min'], -20)
        self.assertEqual(ds2.attrs['geospatial_lon_max'], 60)
        self.assertEqual(ds2.attrs['geospatial_lon_units'], 'degrees_east')
        self.assertEqual(ds2.attrs['geospatial_lon_resolution'], 4)
        self.assertEqual(ds2.attrs['geospatial_bounds'],
                         'POLYGON((-20.0 -42.0, -20.0 42.0, 60.0 42.0, 60.0'
                         ' -42.0, -20.0 -42.0))')

    def test_bnds_inverted(self):
        # Inverted lat
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'second': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'lat': np.linspace(88, -88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': [datetime(2000, x, 1) for x in range(1, 13)]})

        ds.lon.attrs['units'] = 'degrees_east'
        ds.lat.attrs['units'] = 'degrees_north'

        lat_bnds = np.empty([len(ds.lat), 2])
        lon_bnds = np.empty([len(ds.lon), 2])
        ds['nv'] = [0, 1]

        lat_bnds[:, 0] = ds.lat.values + 2
        lat_bnds[:, 1] = ds.lat.values - 2
        lon_bnds[:, 0] = ds.lon.values - 2
        lon_bnds[:, 1] = ds.lon.values + 2

        ds['lat_bnds'] = (['lat', 'nv'], lat_bnds)
        ds['lon_bnds'] = (['lon', 'nv'], lon_bnds)

        ds.lat.attrs['bounds'] = 'lat_bnds'
        ds.lon.attrs['bounds'] = 'lon_bnds'

        ds1 = adjust_spatial_attrs(ds)

        # Make sure original dataset is not altered
        with self.assertRaises(KeyError):
            # noinspection PyStatementEffect
            ds.attrs['geospatial_lat_min']

        # Make sure expected values are in the new dataset
        self.assertEqual(ds1.attrs['geospatial_lat_min'], -90)
        self.assertEqual(ds1.attrs['geospatial_lat_max'], 90)
        self.assertEqual(ds1.attrs['geospatial_lat_units'], 'degrees_north')
        self.assertEqual(ds1.attrs['geospatial_lat_resolution'], 4)
        self.assertEqual(ds1.attrs['geospatial_lon_min'], -180)
        self.assertEqual(ds1.attrs['geospatial_lon_max'], 180)
        self.assertEqual(ds1.attrs['geospatial_lon_units'], 'degrees_east')
        self.assertEqual(ds1.attrs['geospatial_lon_resolution'], 4)
        self.assertEqual(ds1.attrs['geospatial_bounds'],
                         'POLYGON((-180.0 -90.0, -180.0 90.0, 180.0 90.0,'
                         ' 180.0 -90.0, -180.0 -90.0))')

        # Test existing attributes update
        lon_min, lat_min, lon_max, lat_max = -20, -40, 60, 40
        indexers = {'lon': slice(lon_min, lon_max),
                    'lat': slice(lat_max, lat_min)}
        ds2 = ds1.sel(**indexers)
        ds2 = adjust_spatial_attrs(ds2)

        self.assertEqual(ds2.attrs['geospatial_lat_min'], -42)
        self.assertEqual(ds2.attrs['geospatial_lat_max'], 42)
        self.assertEqual(ds2.attrs['geospatial_lat_units'], 'degrees_north')
        self.assertEqual(ds2.attrs['geospatial_lat_resolution'], 4)
        self.assertEqual(ds2.attrs['geospatial_lon_min'], -20)
        self.assertEqual(ds2.attrs['geospatial_lon_max'], 60)
        self.assertEqual(ds2.attrs['geospatial_lon_units'], 'degrees_east')
        self.assertEqual(ds2.attrs['geospatial_lon_resolution'], 4)
        self.assertEqual(
            ds2.attrs['geospatial_bounds'],
            'POLYGON((-20.0 -42.0, -20.0 42.0, 60.0 42.0, '
            '60.0 -42.0, -20.0 -42.0))'
        )

    def test_once_cell_with_bnds(self):
        # Only one cell in lat/lon
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([1, 1, 12])),
            'second': (['lat', 'lon', 'time'], np.zeros([1, 1, 12])),
            'lat': np.array([52.5]),
            'lon': np.array([11.5]),
            'lat_bnds': (['lat', 'bnds'], np.array([[52.4, 52.6]])),
            'lon_bnds': (['lon', 'bnds'], np.array([[11.4, 11.6]])),
            'time': [datetime(2000, x, 1) for x in range(1, 13)]})
        ds.lon.attrs['units'] = 'degrees_east'
        ds.lat.attrs['units'] = 'degrees_north'

        ds1 = adjust_spatial_attrs(ds)
        self.assertAlmostEqual(ds1.attrs['geospatial_lat_resolution'], 0.2)
        self.assertAlmostEqual(ds1.attrs['geospatial_lat_min'], 52.4)
        self.assertAlmostEqual(ds1.attrs['geospatial_lat_max'], 52.6)
        self.assertEqual(ds1.attrs['geospatial_lat_units'], 'degrees_north')
        self.assertAlmostEqual(ds1.attrs['geospatial_lon_resolution'], 0.2)
        self.assertAlmostEqual(ds1.attrs['geospatial_lon_min'], 11.4)
        self.assertAlmostEqual(ds1.attrs['geospatial_lon_max'], 11.6)
        self.assertEqual(ds1.attrs['geospatial_lon_units'], 'degrees_east')
        self.assertEqual(ds1.attrs['geospatial_bounds'],
                         'POLYGON((11.4 52.4, 11.4 52.6, 11.6 52.6, '
                         '11.6 52.4, 11.4 52.4))')

    def test_once_cell_without_bnds(self):
        # Only one cell in lat/lon
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([1, 1, 12])),
            'second': (['lat', 'lon', 'time'], np.zeros([1, 1, 12])),
            'lat': np.array([52.5]),
            'lon': np.array([11.5]),
            'time': [datetime(2000, x, 1) for x in range(1, 13)]})
        ds.lon.attrs['units'] = 'degrees_east'
        ds.lat.attrs['units'] = 'degrees_north'

        ds2 = adjust_spatial_attrs(ds)
        # Datasets should be the same --> not modified
        self.assertIs(ds2, ds)


class AdjustTemporalAttrsTest(TestCase):
    def test_nominal(self):

        time_datas = [
            [datetime(2000, x, 1) for x in range(1, 13)],
            [np.datetime64(datetime(2000, x, 1)) for x in range(1, 13)],
            [cftime.DatetimeGregorian(2000, x, 1) for x in range(1, 13)],
        ]

        for time_data in time_datas:
            print(time_data)
            ds = xr.Dataset({
                'first': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
                'second': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
                'lat': np.linspace(-88, 88, 45),
                'lon': np.linspace(-178, 178, 90),
                'time': time_data
            })

            ds1 = adjust_temporal_attrs(ds)

            # Make sure original dataset is not altered
            with self.assertRaises(KeyError):
                # noinspection PyStatementEffect
                ds.attrs['time_coverage_start']

            # Make sure expected values are in the new dataset
            self.assertEqual(ds1.attrs['time_coverage_start'],
                             '2000-01-01T00:00:00')
            self.assertEqual(ds1.attrs['time_coverage_end'],
                             '2000-12-01T00:00:00')
            self.assertEqual(ds1.attrs['time_coverage_resolution'],
                             'P1M')
            self.assertEqual(ds1.attrs['time_coverage_duration'],
                             'P336D')

            # Test existing attributes update
            # noinspection PyTypeChecker
            indexers = {
                'time': slice(datetime(2000, 2, 15), datetime(2000, 6, 15))
            }
            ds2 = ds1.sel(**indexers)
            ds2 = adjust_temporal_attrs(ds2)

            self.assertEqual(ds2.attrs['time_coverage_start'],
                             '2000-03-01T00:00:00')
            self.assertEqual(ds2.attrs['time_coverage_end'],
                             '2000-06-01T00:00:00')
            self.assertEqual(ds2.attrs['time_coverage_resolution'],
                             'P1M')
            self.assertEqual(ds2.attrs['time_coverage_duration'],
                             'P93D')

    def test_wrong_type(self):
        ds = xr.Dataset({
            'first': (['time', 'lat', 'lon'], np.zeros([12, 45, 90])),
            'second': (['time', 'lat', 'lon'], np.zeros([12, 45, 90])),
            'lon': (['lon'], np.linspace(-178, 178, 90)),
            'lat': (['lat'], np.linspace(-88, 88, 45)),
            'time': (['time'], np.linspace(0, 1, 12))})

        ds1 = adjust_temporal_attrs(ds)

        self.assertIs(ds1, ds)
        self.assertNotIn('time_coverage_start', ds1)
        self.assertNotIn('time_coverage_end', ds1)
        self.assertNotIn('time_coverage_resolution', ds1)
        self.assertNotIn('time_coverage_duration', ds1)

    def test_bnds(self):
        """Test a case when time_bnds is available"""
        time = [datetime(2000, x, 1) for x in range(1, 13)]
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'second': (['lat', 'lon', 'time'], np.zeros([45, 90, 12])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'nv': [0, 1],
            'time': time})

        month_ends = list()
        for x in ds.time.values:
            year = int(str(x)[0:4])
            month = int(str(x)[5:7])
            day = calendar.monthrange(year, month)[1]
            month_ends.append(datetime(year, month, day))

        ds['time_bnds'] = (['time', 'nv'], list(zip(time, month_ends)))
        ds.time.attrs['bounds'] = 'time_bnds'

        ds1 = adjust_temporal_attrs(ds)

        # Make sure original dataset is not altered
        with self.assertRaises(KeyError):
            # noinspection PyStatementEffect
            ds.attrs['time_coverage_start']

        # Make sure expected values are in the new dataset
        self.assertEqual(ds1.attrs['time_coverage_start'],
                         '2000-01-01T00:00:00')
        self.assertEqual(ds1.attrs['time_coverage_end'],
                         '2000-12-31T00:00:00')
        self.assertEqual(ds1.attrs['time_coverage_resolution'],
                         'P1M')
        self.assertEqual(ds1.attrs['time_coverage_duration'],
                         'P366D')

    def test_single_slice(self):
        """Test a case when the dataset is a single time slice"""
        # With bnds
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([45, 90, 1])),
            'second': (['lat', 'lon', 'time'], np.zeros([45, 90, 1])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'nv': [0, 1],
            'time': [datetime(2000, 1, 1)]})
        ds.time.attrs['bounds'] = 'time_bnds'
        ds['time_bnds'] = (['time', 'nv'],
                           [(datetime(2000, 1, 1), datetime(2000, 1, 31))])

        ds1 = adjust_temporal_attrs(ds)

        # Make sure original dataset is not altered
        with self.assertRaises(KeyError):
            # noinspection PyStatementEffect
            ds.attrs['time_coverage_start']

        # Make sure expected values are in the new dataset
        self.assertEqual(ds1.attrs['time_coverage_start'],
                         '2000-01-01T00:00:00')
        self.assertEqual(ds1.attrs['time_coverage_end'],
                         '2000-01-31T00:00:00')
        self.assertEqual(ds1.attrs['time_coverage_duration'],
                         'P31D')
        with self.assertRaises(KeyError):
            # Resolution is not defined for a single slice
            # noinspection PyStatementEffect
            ds.attrs['time_coverage_resolution']

        # Without bnds
        ds = xr.Dataset({
            'first': (['lat', 'lon', 'time'], np.zeros([45, 90, 1])),
            'second': (['lat', 'lon', 'time'], np.zeros([45, 90, 1])),
            'lat': np.linspace(-88, 88, 45),
            'lon': np.linspace(-178, 178, 90),
            'time': [datetime(2000, 1, 1)]})

        ds1 = adjust_temporal_attrs(ds)

        self.assertEqual(ds1.attrs['time_coverage_start'],
                         '2000-01-01T00:00:00')
        self.assertEqual(ds1.attrs['time_coverage_end'],
                         '2000-01-01T00:00:00')
        with self.assertRaises(KeyError):
            # noinspection PyStatementEffect
            ds.attrs['time_coverage_resolution']
        with self.assertRaises(KeyError):
            # noinspection PyStatementEffect
            ds.attrs['time_coverage_duration']


class TestNormalizeCoordVars(TestCase):

    def test_ds_with_potential_coords(self):
        ds = xr.Dataset({'first': (['lat', 'lon'], np.zeros([90, 180])),
                         'second': (['lat', 'lon'], np.zeros([90, 180])),
                         'lat_bnds': (['lat', 'bnds'], np.zeros([90, 2])),
                         'lon_bnds': (['lon', 'bnds'], np.zeros([180, 2]))},
                        coords={'lat': np.linspace(-89.5, 89.5, 90),
                                'lon': np.linspace(-179.5, 179.5, 180)})

        new_ds = normalize_coord_vars(ds)

        self.assertIsNot(ds, new_ds)
        self.assertEqual(len(new_ds.coords), 4)
        self.assertIn('lon', new_ds.coords)
        self.assertIn('lat', new_ds.coords)
        self.assertIn('lat_bnds', new_ds.coords)
        self.assertIn('lon_bnds', new_ds.coords)

        self.assertEqual(len(new_ds.data_vars), 2)
        self.assertIn('first', new_ds.data_vars)
        self.assertIn('second', new_ds.data_vars)

    def test_ds_with_potential_coords_and_bounds(self):
        ds = xr.Dataset({'first': (['lat', 'lon'], np.zeros([90, 180])),
                         'second': (['lat', 'lon'], np.zeros([90, 180])),
                         'lat_bnds': (['lat', 'bnds'], np.zeros([90, 2])),
                         'lon_bnds': (['lon', 'bnds'], np.zeros([180, 2])),
                         'lat': (['lat'], np.linspace(-89.5, 89.5, 90)),
                         'lon': (['lon'], np.linspace(-179.5, 179.5, 180))})

        new_ds = normalize_coord_vars(ds)

        self.assertIsNot(ds, new_ds)
        self.assertEqual(len(new_ds.coords), 4)
        self.assertIn('lon', new_ds.coords)
        self.assertIn('lat', new_ds.coords)
        self.assertIn('lat_bnds', new_ds.coords)
        self.assertIn('lon_bnds', new_ds.coords)

        self.assertEqual(len(new_ds.data_vars), 2)
        self.assertIn('first', new_ds.data_vars)
        self.assertIn('second', new_ds.data_vars)

    def test_ds_with_no_potential_coords(self):
        ds = xr.Dataset({'first': (['lat', 'lon'], np.zeros([90, 180])),
                         'second': (['lat', 'lon'], np.zeros([90, 180]))},
                        coords={'lat': np.linspace(-89.5, 89.5, 90),
                                'lon': np.linspace(-179.5, 179.5, 180)},
                        attrs={'time_coverage_start': '20120101'})
        new_ds = normalize_coord_vars(ds)
        self.assertIs(ds, new_ds)


class TestNormalizeMissingTime(TestCase):
    def test_ds_without_time(self):
        ds = xr.Dataset({'first': (['lat', 'lon'], np.zeros([90, 180])),
                         'second': (['lat', 'lon'], np.zeros([90, 180]))},
                        coords={'lat': np.linspace(-89.5, 89.5, 90),
                                'lon': np.linspace(-179.5, 179.5, 180)},
                        attrs={'time_coverage_start': '20120101',
                               'time_coverage_end': '20121231'})

        new_ds = normalize_missing_time(ds)

        self.assertIsNot(ds, new_ds)
        self.assertEqual(len(new_ds.coords), 4)
        self.assertIn('lon', new_ds.coords)
        self.assertIn('lat', new_ds.coords)
        self.assertIn('time', new_ds.coords)
        self.assertIn('time_bnds', new_ds.coords)

        self.assertEqual(new_ds.coords['time'].attrs.get('long_name'), 'time')
        self.assertEqual(new_ds.coords['time'].attrs.get('bounds'), 'time_bnds')

        self.assertEqual(new_ds.first.shape, (1, 90, 180))
        self.assertEqual(new_ds.second.shape, (1, 90, 180))
        self.assertEqual(new_ds.coords['time'][0],
                         xr.DataArray(pd.to_datetime('2012-07-01T12:00:00')))
        self.assertEqual(new_ds.coords['time'].attrs.get('long_name'), 'time')
        self.assertEqual(new_ds.coords['time'].attrs.get('bounds'), 'time_bnds')
        self.assertEqual(new_ds.coords['time_bnds'][0][0],
                         xr.DataArray(pd.to_datetime('2012-01-01')))
        self.assertEqual(new_ds.coords['time_bnds'][0][1],
                         xr.DataArray(pd.to_datetime('2012-12-31')))
        self.assertEqual(
            new_ds.coords['time_bnds'].attrs.get('long_name'), 'time'
        )

    def test_ds_without_bounds(self):
        ds = xr.Dataset({'first': (['lat', 'lon'], np.zeros([90, 180])),
                         'second': (['lat', 'lon'], np.zeros([90, 180]))},
                        coords={'lat': np.linspace(-89.5, 89.5, 90),
                                'lon': np.linspace(-179.5, 179.5, 180)},
                        attrs={'time_coverage_start': '20120101'})

        new_ds = normalize_missing_time(ds)

        self.assertIsNot(ds, new_ds)
        self.assertEqual(len(new_ds.coords), 3)
        self.assertIn('lon', new_ds.coords)
        self.assertIn('lat', new_ds.coords)
        self.assertIn('time', new_ds.coords)
        self.assertNotIn('time_bnds', new_ds.coords)

        self.assertEqual(new_ds.first.shape, (1, 90, 180))
        self.assertEqual(new_ds.second.shape, (1, 90, 180))
        self.assertEqual(
            new_ds.coords['time'][0], xr.DataArray(pd.to_datetime('2012-01-01'))
        )
        self.assertEqual(new_ds.coords['time'].attrs.get('long_name'), 'time')
        self.assertEqual(new_ds.coords['time'].attrs.get('bounds'), None)

    def test_ds_without_time_attrs(self):
        ds = xr.Dataset({'first': (['lat', 'lon'], np.zeros([90, 180])),
                         'second': (['lat', 'lon'], np.zeros([90, 180]))},
                        coords={'lat': np.linspace(-89.5, 89.5, 90),
                                'lon': np.linspace(-179.5, 179.5, 180)})

        new_ds = normalize_missing_time(ds)
        self.assertIs(ds, new_ds)

    def test_ds_with_cftime(self):
        time_data = xr.cftime_range(start='2010-01-01T00:00:00',
                                    periods=6,
                                    freq='D',
                                    calendar='gregorian').values
        ds = xr.Dataset(
            {'first': (['time', 'lat', 'lon'], np.zeros([6, 90, 180])),
             'second': (['time', 'lat', 'lon'], np.zeros([6, 90, 180]))},
            coords={'lat': np.linspace(-89.5, 89.5, 90),
                    'lon': np.linspace(-179.5, 179.5, 180),
                    'time': time_data},
            attrs={'time_coverage_start': '20120101',
                   'time_coverage_end': '20121231'}
        )
        new_ds = normalize_missing_time(ds)
        self.assertIs(ds, new_ds)


class Fix360Test(TestCase):

    def test_fix_360_lon(self):
        # The following simulates a strangely geo-coded soil moisture dataset
        # we found at ...
        #
        lon_size = 360
        lat_size = 130
        time_size = 12
        ds = xr.Dataset({
            'first': (['time', 'lat', 'lon'],
                      np.random.random_sample([time_size, lat_size, lon_size])),
            'second': (['time', 'lat', 'lon'],
                       np.random.random_sample([time_size, lat_size, lon_size]))
        },
            coords={'lon': np.linspace(1., 360., lon_size),
                    'lat': np.linspace(-64., 65., lat_size),
                    'time':
                        [datetime(2000, x, 1) for x in range(1, time_size + 1)]
                    },
            attrs=dict(geospatial_lon_min=0.,
                       geospatial_lon_max=360.,
                       geospatial_lat_min=-64.5,
                       geospatial_lat_max=+65.5,
                       geospatial_lon_resolution=1.,
                       geospatial_lat_resolution=1.))

        new_ds = normalize(ds)
        self.assertIsNot(ds, new_ds)
        self.assertEqual(ds.dims, new_ds.dims)
        self.assertEqual(ds.sizes, new_ds.sizes)
        assert_array_almost_equal(new_ds.lon, np.linspace(-179.5, 179.5, 360))
        assert_array_almost_equal(new_ds.lat, np.linspace(-64., 65., 130))
        assert_array_almost_equal(new_ds.first[..., :180], ds.first[..., 180:])
        assert_array_almost_equal(new_ds.first[..., 180:], ds.first[..., :180])
        assert_array_almost_equal(new_ds.second[..., :180],
                                  ds.second[..., 180:])
        assert_array_almost_equal(new_ds.second[..., 180:],
                                  ds.second[..., :180])
        self.assertEqual(-180., new_ds.attrs['geospatial_lon_min'])
        self.assertEqual(+180., new_ds.attrs['geospatial_lon_max'])
        self.assertEqual(-64.5, new_ds.attrs['geospatial_lat_min'])
        self.assertEqual(+65.5, new_ds.attrs['geospatial_lat_max'])
        self.assertEqual(1., new_ds.attrs['geospatial_lon_resolution'])
        self.assertEqual(1., new_ds.attrs['geospatial_lat_resolution'])


class NormalizeDimOrder(TestCase):
    """
    Test normalize_cci_sea_level operation
    """

    def test_no_change(self):
        """
        Test nominal operation
        """
        lon_size = 360
        lat_size = 130
        time_size = 12
        ds = xr.Dataset(
            {
                'first': (
                    ['time', 'lat', 'lon'],
                    np.random.random_sample([time_size, lat_size, lon_size])),
                'second': (
                    ['time', 'lat', 'lon'],
                    np.random.random_sample([time_size, lat_size, lon_size]))
            },
            coords={
                'lon': np.linspace(-179.5, -179.5, lon_size),
                'lat': np.linspace(-64., 65., lat_size),
                'time': [datetime(2000, x, 1) for x in range(1, time_size + 1)]
            }
        )
        ds2 = normalize(ds)
        self.assertIs(ds2, ds)

    def test_nominal(self):
        """
        Test nominal operation
        """
        ds = self.new_cci_sea_level_ds()
        ds2 = normalize(ds)
        self.assertIsNot(ds2, ds)
        self.assertIn('ampl', ds2)
        self.assertIn('phase', ds2)
        self.assertIn('time', ds2.coords)
        self.assertIn('time_bnds', ds2.coords)
        self.assertNotIn('time_step', ds2.coords)
        self.assertEqual(['time', 'period', 'lat', 'lon'], list(ds2.ampl.dims))
        self.assertEqual(['time', 'period', 'lat', 'lon'], list(ds2.phase.dims))

    def new_cci_sea_level_ds(self):
        period_size = 2
        lon_size = 4
        lat_size = 2

        dataset = xr.Dataset(dict(
            ampl=(['lat', 'lon', 'period'],
                  np.ones(shape=(lat_size, lon_size, period_size))),
            phase=(['lat', 'lon', 'period'],
                   np.zeros(shape=(lat_size, lon_size, period_size)))),
            coords=dict(
                lon=np.array([-135, -45., 45., 135.]),
                lat=np.array([-45., 45.]),
                time=pd.to_datetime(
                    ['1993-01-15T00:00:00.000000000',
                     '1993-02-15T00:00:00.000000000',
                     '2015-11-15T00:00:00.000000000',
                     '2015-12-15T00:00:00.000000000'])
            )
        )

        dataset.coords['time'].encoding.update(
            units='days since 1950-01-01',
            dtype=np.dtype(np.float32)
        )
        dataset.coords['time'].attrs.update(
            long_name='time',
            standard_name='time'
        )

        dataset.attrs['time_coverage_start'] = '1993-01-01 00:00:00'
        dataset.attrs['time_coverage_end'] = '2015-12-31 23:59:59'

        return dataset



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


__author__ = "Norman Fomferra (Brockmann Consult GmbH)"

import atexit
import logging
import os
import tempfile

_TEMP_FILES = []

_LOG = logging.getLogger('esa-climate-toolbox')


def new_temp_file(**kwargs):
    """
    Delegates to standard Python ``tempfile.mkstemp(**kwargs)`` function and
    stores the returned pair in a global list files to be removed when the
    interpreter shuts down.
    :param kwargs: Same as for ``tempfile.mkstemp()``
    :return: a pair (fd, file) as returned by ``tempfile.mkstemp()``
    """
    pair = tempfile.mkstemp(**kwargs)
    _TEMP_FILES.append(pair)
    return pair


def del_temp_file(file):
    """
    Removes an individual file or pair (fd, file) registered
    by :py:func:`new_temp_file`.
    :param file: The file path or pair (fd, file) as returned
        by :py:func:`new_temp_file`.
    """
    global _TEMP_FILES

    try:
        _, file = file
    except ValueError:
        pass

    for i in range(len(_TEMP_FILES)):
        pair = _TEMP_FILES[i]
        if os.path.samefile(pair[1], file):
            is_deleted = _del_file(pair)
            if is_deleted:
                del _TEMP_FILES[i]
            return is_deleted
    return False


def del_temp_files(force=False):
    """
    Removes all temporary files registered by :py:func:`new_temp_file`.
    This function is called as a shutdown hook of the Python interpreter.
    It may be called directly if temporary files consume lots of disk space.
    """
    global _TEMP_FILES
    remaining = list()
    for pair in list(_TEMP_FILES):
        if not _del_file(pair):
            remaining.append(pair)
    _TEMP_FILES = [] if force else remaining


# Register del_temp_files() as a shutdown hook.
atexit.register(del_temp_files)


def _del_file(pair: str) -> bool:
    fd, file = pair
    if os.path.isfile(file):
        try:
            os.close(fd)
        except OSError:
            pass
        try:
            os.remove(file)
        except OSError:
            _LOG.exception('removing temporary file %s failed' % file)
        return not os.path.isfile(file)
    else:
        return True


def get_temp_files():
    """
    Return the list of temporary files as list of pairs (fd, file).
    For testing only.
    """
    global _TEMP_FILES
    return _TEMP_FILES



# The MIT License (MIT)
# Copyright (c) 2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Description
===========

Data animation operations

Components
==========

## animate_map

Animates a geospatial data slice on a world map.

General jupyter notebook usage:
```python
import esa_climate_toolbox.ops as ops
from IPython.core.display import display, HTML
dset = ops.open_dataset('local.dataset_name')
display(HTML(ops.animate_map(cc, var='var_name')))
```

If a file path is given, the plot is saved.
Supported formats: html

"""

import os

import matplotlib
import matplotlib.animation as animation
import matplotlib.pyplot as plt

import cartopy.crs as ccrs
import xarray as xr
import numpy as np

from esa_climate_toolbox.core.op import op
from esa_climate_toolbox.core.op import op_input
from esa_climate_toolbox.core.types import DictLike
from esa_climate_toolbox.core.types import HTML
from esa_climate_toolbox.core.types import PolygonLike
from esa_climate_toolbox.core.types import ValidationError
from esa_climate_toolbox.core.types import VarName
from esa_climate_toolbox.util.monitor import Monitor

from esa_climate_toolbox.ops.plot_helpers import determine_cmap_params
from esa_climate_toolbox.ops.plot_helpers import get_var_data
from esa_climate_toolbox.ops.plot_helpers import handle_plot_polygon

matplotlib.use("Agg")

ANIMATION_FILE_FILTER = dict(name='Animation Outputs', extensions=['html', ])


@op(tags=['plot'], res_pattern='animation_{index}')
@op_input('ds')
@op_input('var', value_set_source='ds', data_type=VarName)
@op_input('indexers', data_type=DictLike)
@op_input('region', data_type=PolygonLike)
@op_input('projection', value_set=[
    'PlateCarree', 'LambertCylindrical', 'Mercator', 'Miller', 'Mollweide',
    'Orthographic', 'Robinson', 'Sinusoidal', 'NorthPolarStereo', 'SouthPolarStereo']
          )
@op_input('central_lon', units='degrees', value_range=[-180, 180])
@op_input('title')
@op_input('cmap_params', data_type=DictLike)
@op_input('plot_properties', data_type=DictLike)
@op_input('file', file_open_mode='w', file_filters=[ANIMATION_FILE_FILTER])
def animate_map(ds: xr.Dataset,
                var: VarName.TYPE = None,
                animate_dim: str = 'time',
                interval: int = 200,
                true_range: bool = False,
                indexers: DictLike.TYPE = None,
                region: PolygonLike.TYPE = None,
                projection: str = 'PlateCarree',
                central_lon: float = 0.0,
                title: str = None,
                contour_plot: bool = False,
                cmap_params: DictLike.TYPE = None,
                plot_properties: DictLike.TYPE = None,
                file: str = None,
                monitor: Monitor = Monitor.NONE) -> HTML:
    """
    Create a geographic map animation for the variable given by dataset *ds*
    and variable name *var*.

    Creates an animation of the given variable from the given dataset on a map
    with coastal lines.
    In case no variable name is given, the first encountered variable in the
    dataset is animated.
    It is also possible to set extents of the animation. If no extents
    are given, a global animation is created.

    The following file formats for saving the animation are supported: html

    :param ds: the dataset containing the variable to animate
    :param var: the variable's name
    :param animate_dim: Dimension to animate, if none given defaults to time.
    :param interval: Delay between frames in milliseconds. Defaults to 200.
    :param true_range: If True, calculates colormap and colorbar configuration
        parameters from the whole dataset. Can potentially take a lot of time.
        Defaults to False, in which case the colormap is calculated from the first frame
    :param indexers: Optional indexers into data array of *var*. The *indexers* is a
        dictionary or a comma-separated string of key-value pairs that maps the
        variable's dimension names to constant labels. e.g. "layer=4".
    :param region: Region to animate
    :param projection: name of a global projection,
        see http://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html
    :param central_lon: central longitude of the projection in degrees
    :param title: an optional title
    :param contour_plot: If true plot a filled contour plot of data,
        otherwise plots a pixelated colormesh
    :param cmap_params: optional additional colormap configuration parameters,
           e.g. "vmax=300, cmap='magma'"
           For full reference refer to
           http://xarray.pydata.org/en/stable/generated/xarray.plot.contourf.html
    :param plot_properties: optional plot properties for Python matplotlib,
           e.g. "bins=512, range=(-1.5, +1.5)"
           For full reference refer to
           https://matplotlib.org/api/lines_api.html and
           https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.contourf.html
    :param file: path to a file in which to save the animation
    :param monitor: A progress monitor.
    :return: An animation in HTML format
    """
    if not isinstance(ds, xr.Dataset):
        raise NotImplementedError('Only gridded datasets are currently supported')

    var_name = None
    if not var:
        for key in ds.data_vars.keys():
            var_name = key
            break
    else:
        var_name = VarName.convert(var)

    try:
        var = ds[var_name]
    except KeyError:
        raise ValidationError(
            'Provided variable name "{}" does not exist '
            'in the given dataset'.format(var_name)
        )

    indexers = DictLike.convert(indexers) or {}
    properties = DictLike.convert(plot_properties) or {}
    cmap_params = DictLike.convert(cmap_params) or {}

    extents = None
    bounds = handle_plot_polygon(region)
    if bounds:
        lon_min, lat_min, lon_max, lat_max = bounds
        extents = [lon_min, lon_max, lat_min, lat_max]

    if len(ds.lat) < 2 or len(ds.lon) < 2:
        # Matplotlib can not plot datasets with less than these dimensions with
        # contourf and pcolormesh methods
        raise ValidationError('The minimum dataset spatial dimensions to create a map'
                              ' plot are (2,2)')

    # See http://scitools.org.uk/cartopy/docs/v0.15/crs/projections.html#
    if projection == 'PlateCarree':
        proj = ccrs.PlateCarree(central_longitude=central_lon)
    elif projection == 'LambertCylindrical':
        proj = ccrs.LambertCylindrical(central_longitude=central_lon)
    elif projection == 'Mercator':
        proj = ccrs.Mercator(central_longitude=central_lon)
    elif projection == 'Miller':
        proj = ccrs.Miller(central_longitude=central_lon)
    elif projection == 'Mollweide':
        proj = ccrs.Mollweide(central_longitude=central_lon)
    elif projection == 'Orthographic':
        proj = ccrs.Orthographic(central_longitude=central_lon)
    elif projection == 'Robinson':
        proj = ccrs.Robinson(central_longitude=central_lon)
    elif projection == 'Sinusoidal':
        proj = ccrs.Sinusoidal(central_longitude=central_lon)
    elif projection == 'NorthPolarStereo':
        proj = ccrs.NorthPolarStereo(central_longitude=central_lon)
    elif projection == 'SouthPolarStereo':
        proj = ccrs.SouthPolarStereo(central_longitude=central_lon)
    else:
        raise ValidationError('illegal projection: "%s"' % projection)

    figure = plt.figure(figsize=(8, 4))
    ax = plt.axes(projection=proj)
    if extents:
        ax.set_extent(extents, ccrs.PlateCarree())
    else:
        ax.set_global()

    ax.coastlines()

    if not animate_dim:
        animate_dim = 'time'

    indexers[animate_dim] = var[animate_dim][0]

    var_data = get_var_data(var, indexers, remaining_dims=('lon', 'lat'))

    with monitor.starting("animate", len(var[animate_dim]) + 3):
        if true_range:
            data_min, data_max = _get_min_max(var, monitor=monitor)
        else:
            data_min, data_max = _get_min_max(var_data, monitor=monitor)

        cmap_params = determine_cmap_params(data_min, data_max, **cmap_params)
        plot_kwargs = {**properties, **cmap_params}

        # Plot the first frame to set-up the axes with the colorbar properly
        # transform keyword is for the coordinate our data is in, which in case of a
        # 'normal' lat/lon dataset is PlateCarree.
        if contour_plot:
            var_data.plot.contourf(
                ax=ax, transform=ccrs.PlateCarree(), add_colorbar=True, **plot_kwargs
            )
        else:
            var_data.plot.pcolormesh(
                ax=ax, transform=ccrs.PlateCarree(), add_colorbar=True, **plot_kwargs
            )
        if title:
            ax.set_title(title)
        figure.tight_layout()
        monitor.progress(1)

        def run(value):
            ax.clear()
            if extents:
                ax.set_extent(extents, ccrs.PlateCarree())
            else:
                ax.set_global()
            ax.coastlines()
            indexers[animate_dim] = value
            indexed_var_data = get_var_data(
                var, indexers, remaining_dims=('lon', 'lat')
            )
            indexed_var_data.plot.contourf(ax=ax, transform=ccrs.PlateCarree(),
                                           add_colorbar=False, **plot_kwargs)
            if title:
                ax.set_title(title)
            monitor.progress(1)
            return ax

        # Reports Optional[int] expected but iterable allowed
        # noinspection PyTypeChecker
        anim = animation.FuncAnimation(figure, run, [i for i in var[animate_dim]],
                                       interval=interval, blit=False, repeat=False)
        anim_html = anim.to_jshtml()

        # Prevent the animation for running after it's finished
        del anim

        # Delete the rogue temp-file
        try:
            os.remove('None0000000.png')
        except FileNotFoundError:
            pass

        if file:
            with open(file, 'w') as outfile:
                outfile.write(anim_html)
                monitor.progress(1)

    return HTML(anim_html)


def _get_min_max(data, monitor=None):
    """
    Get min and max of a dataset, while accounting for all-NaN
    datasets and observing it with the monitor.
    """
    with monitor.child(1).observing("find minimum"):
        data_min = data.min()
    if np.isnan(data_min):
        # Handle all-NaN dataset
        raise ValidationError(
            'Can not create an animation of a dataset containing only NaN values.'
        )
    else:
        with monitor.child(1).observing("find maximum"):
            data_max = data.max()

    return data_min.values, data_max.values
